"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import os
import re
import unittest
from vgrid.utils.eaggr.eaggr import Eaggr
from vgrid.utils.eaggr.enums.model import Model
from vgrid.utils.eaggr.enums.dggs_shape_type import DggsShapeType
from vgrid.utils.eaggr.enums.dggs_shape_location import DggsShapeLocation
from vgrid.utils.eaggr.enums.shape_string_format import ShapeStringFormat
from vgrid.utils.eaggr.shapes.lat_long_point import LatLongPoint
from vgrid.utils.eaggr.shapes.lat_long_linestring import LatLongLinestring
from vgrid.utils.eaggr.shapes.lat_long_polygon import LatLongPolygon
from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
from vgrid.utils.eaggr.shapes.dggs_linestring import DggsLinestring
from vgrid.utils.eaggr.shapes.dggs_polygon import DggsPolygon
from vgrid.utils.eaggr.enums.dggs_analysis_type import DggsAnalysisType
from vgrid.utils.eaggr.shapes.dggs_shape import DggsShape


## Sets up test data for ISEA4T spatial analysis unit tests
class DggsTestDataTriangular:
    # Initialise the base shapes for spatial analysis testing on a triangular grid
    # Create a cell
    dggs_shape_cell = DggsShape(DggsCell("0700"), DggsShapeLocation.ONE_FACE)
    # Create a linestring
    dggs_linestring = DggsLinestring()
    dggs_linestring.add_cell(DggsCell("0701"))
    dggs_linestring.add_cell(DggsCell("0702"))
    dggs_linestring.add_cell(DggsCell("0703"))
    dggs_shape_linestring = DggsShape(dggs_linestring, DggsShapeLocation.ONE_FACE)
    # Create a polygon with an inner ring
    dggs_outer_linestring = DggsLinestring()
    dggs_outer_linestring.add_cell(DggsCell("07111"))
    dggs_outer_linestring.add_cell(DggsCell("0702"))
    dggs_outer_linestring.add_cell(DggsCell("0701"))
    dggs_outer_linestring.add_cell(DggsCell("07333"))
    dggs_polygon = DggsPolygon(dggs_outer_linestring)
    dggs_inner_linestring = DggsLinestring()
    dggs_inner_linestring.add_cell(DggsCell("07312"))
    dggs_inner_linestring.add_cell(DggsCell("07302"))
    dggs_inner_linestring.add_cell(DggsCell("07321"))
    dggs_polygon.add_inner_ring(dggs_inner_linestring)
    dggs_polygon = DggsPolygon(dggs_outer_linestring)
    dggs_shape_polygon = DggsShape(dggs_polygon, DggsShapeLocation.ONE_FACE)
    # Initialise comparison shapes from cells, for spatial analysis testing
    # Create a eaggr shape from a child cell of the base cell
    child_dggs_shape_cell = DggsShape(DggsCell("07000"), DggsShapeLocation.ONE_FACE)
    # Create a eaggr shape from a parent cell of the base cell
    parent_dggs_shape_cell = DggsShape(DggsCell("070"), DggsShapeLocation.ONE_FACE)
    # Create a eaggr shape from a cell that is disjoint from the base cell
    disjoint_dggs_shape_cell = DggsShape(DggsCell("0711"), DggsShapeLocation.ONE_FACE)
    # Create a eaggr shape from a cell that shares an edge with the base cell
    edge_sharing_dggs_shape_cell = DggsShape(
        DggsCell("0701"), DggsShapeLocation.ONE_FACE
    )
    # Initialise comparison shapes from linestrings, for spatial analysis testing
    # Create a eaggr shape from a linestring that is a substring of the base linestring
    substring_dggs_linestring = DggsLinestring()
    substring_dggs_linestring.add_cell(DggsCell("0701"))
    substring_dggs_linestring.add_cell(DggsCell("0702"))
    substring_dggs_shape_linestring = DggsShape(
        substring_dggs_linestring, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a linestring that is disjoint from the base linestring
    disjoint_dggs_linestring = DggsLinestring()
    disjoint_dggs_linestring.add_cell(DggsCell("0721"))
    disjoint_dggs_linestring.add_cell(DggsCell("0722"))
    disjoint_dggs_linestring.add_cell(DggsCell("0723"))
    disjoint_dggs_shape_linestring = DggsShape(
        disjoint_dggs_linestring, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a linestring that is a superstring of the base linestring
    superstring_dggs_linestring = DggsLinestring()
    superstring_dggs_linestring.add_cell(DggsCell("0733"))
    superstring_dggs_linestring.add_cell(DggsCell("0701"))
    superstring_dggs_linestring.add_cell(DggsCell("0702"))
    superstring_dggs_linestring.add_cell(DggsCell("0703"))
    superstring_dggs_linestring.add_cell(DggsCell("0711"))
    superstring_dggs_shape_linestring = DggsShape(
        superstring_dggs_linestring, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a linestring whose end touches the end of the base linestring
    ends_touching_dggs_linestring = DggsLinestring()
    ends_touching_dggs_linestring.add_cell(DggsCell("0733"))
    ends_touching_dggs_linestring.add_cell(DggsCell("0732"))
    ends_touching_dggs_linestring.add_cell(DggsCell("0701"))
    ends_touching_dggs_shape_linestring = DggsShape(
        ends_touching_dggs_linestring, DggsShapeLocation.ONE_FACE
    )
    # Initialise comparison shapes from polygons, for spatial analysis testing
    # Create a eaggr shape from a polygon with no inner rings, that lies within the interior of the base polygon
    interior_polygon_dggs_outer_linestring = DggsLinestring()
    interior_polygon_dggs_outer_linestring.add_cell(DggsCell("07101"))
    interior_polygon_dggs_outer_linestring.add_cell(DggsCell("07132"))
    interior_polygon_dggs_outer_linestring.add_cell(DggsCell("07001"))
    interior_polygon_dggs_polygon = DggsPolygon(interior_polygon_dggs_outer_linestring)
    interior_polygon_dggs_shape_polygon = DggsShape(
        interior_polygon_dggs_polygon, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a polygon with no inner rings, that is disjoint from the base polygon
    disjoint_polygon_dggs_outer_linestring = DggsLinestring()
    disjoint_polygon_dggs_outer_linestring.add_cell(DggsCell("0721"))
    disjoint_polygon_dggs_outer_linestring.add_cell(DggsCell("0722"))
    disjoint_polygon_dggs_outer_linestring.add_cell(DggsCell("0723"))
    disjoint_polygon_dggs_polygon = DggsPolygon(disjoint_polygon_dggs_outer_linestring)
    disjoint_polygon_dggs_shape_polygon = DggsShape(
        disjoint_polygon_dggs_polygon, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a polygon with no inner rings, that contains the base polygon
    contains_polygon_dggs_outer_linestring = DggsLinestring()
    contains_polygon_dggs_outer_linestring.add_cell(DggsCell("07111"))
    contains_polygon_dggs_outer_linestring.add_cell(DggsCell("07222"))
    contains_polygon_dggs_outer_linestring.add_cell(DggsCell("07333"))
    contains_polygon_dggs_polygon = DggsPolygon(contains_polygon_dggs_outer_linestring)
    contains_polygon_dggs_shape_polygon = DggsShape(
        contains_polygon_dggs_polygon, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a polygon with no inner rings, that shares an edge with the base polygon
    edge_sharing_polygon_dggs_outer_linestring = DggsLinestring()
    edge_sharing_polygon_dggs_outer_linestring.add_cell(DggsCell("0702"))
    edge_sharing_polygon_dggs_outer_linestring.add_cell(DggsCell("0701"))
    edge_sharing_polygon_dggs_outer_linestring.add_cell(DggsCell("0722"))
    edge_sharing_polygon_dggs_polygon = DggsPolygon(
        edge_sharing_polygon_dggs_outer_linestring
    )
    edge_sharing_polygon_dggs_shape_polygon = DggsShape(
        edge_sharing_polygon_dggs_polygon, DggsShapeLocation.ONE_FACE
    )


## Sets up test data for ISEA4T spatial analysis unit tests
class DggsTestDataHexagonal:
    # Initialise the base shapes for spatial analysis testing on a triangular grid
    # Create a cell
    dggs_shape_cell = DggsShape(DggsCell("07021,1"), DggsShapeLocation.ONE_FACE)
    # Create a linestring
    dggs_linestring = DggsLinestring()
    dggs_linestring.add_cell(DggsCell("07021,2"))
    dggs_linestring.add_cell(DggsCell("07023,1"))
    dggs_linestring.add_cell(DggsCell("07023,3"))
    dggs_shape_linestring = DggsShape(dggs_linestring, DggsShapeLocation.ONE_FACE)
    # Create a polygon with an inner ring
    dggs_outer_linestring = DggsLinestring()
    dggs_outer_linestring.add_cell(DggsCell("07020,2"))
    dggs_outer_linestring.add_cell(DggsCell("07024,0"))
    dggs_outer_linestring.add_cell(DggsCell("07024,4"))
    dggs_outer_linestring.add_cell(DggsCell("07023,4"))
    dggs_outer_linestring.add_cell(DggsCell("07020,3"))
    dggs_polygon = DggsPolygon(dggs_outer_linestring)
    dggs_inner_linestring = DggsLinestring()
    dggs_inner_linestring.add_cell(DggsCell("07021,2"))
    dggs_inner_linestring.add_cell(DggsCell("07023,1"))
    dggs_inner_linestring.add_cell(DggsCell("07023,3"))
    dggs_polygon.add_inner_ring(dggs_inner_linestring)
    dggs_shape_polygon = DggsShape(dggs_polygon, DggsShapeLocation.ONE_FACE)
    # Initialise comparison shapes from cells, for spatial analysis testing
    # Create a eaggr shape from a partially overlapping child cell of the base cell
    overlapping_child_dggs_shape_cell = DggsShape(
        DggsCell("07031,4"), DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a cell that is disjoint from the base cell
    disjoint_dggs_shape_cell = DggsShape(
        DggsCell("07023,1"), DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a linestring that is disjoint from the base linestring
    disjoint_dggs_linestring = DggsLinestring()
    disjoint_dggs_linestring.add_cell(DggsCell("07020,2"))
    disjoint_dggs_linestring.add_cell(DggsCell("07023,0"))
    disjoint_dggs_linestring.add_cell(DggsCell("07024,1"))
    disjoint_dggs_linestring.add_cell(DggsCell("07024,4"))
    disjoint_dggs_shape_linestring = DggsShape(
        disjoint_dggs_linestring, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a polygon with no inner rings, that overlaps the base polygon
    overlapping_polygon_dggs_outer_linestring = DggsLinestring()
    overlapping_polygon_dggs_outer_linestring.add_cell(DggsCell("07032,3"))
    overlapping_polygon_dggs_outer_linestring.add_cell(DggsCell("07033,4"))
    overlapping_polygon_dggs_outer_linestring.add_cell(DggsCell("07022,3"))
    overlapping_polygon_dggs_polygon = DggsPolygon(
        overlapping_polygon_dggs_outer_linestring
    )
    overlapping_polygon_dggs_shape_polygon = DggsShape(
        overlapping_polygon_dggs_polygon, DggsShapeLocation.ONE_FACE
    )
    # Create a eaggr shape from a polygon with no inner rings, that is disjoint from the base polygon
    disjoint_polygon_dggs_outer_linestring = DggsLinestring()
    disjoint_polygon_dggs_outer_linestring.add_cell(DggsCell("07020,0"))
    disjoint_polygon_dggs_outer_linestring.add_cell(DggsCell("07020,1"))
    disjoint_polygon_dggs_outer_linestring.add_cell(DggsCell("07021,0"))
    disjoint_polygon_dggs_polygon = DggsPolygon(disjoint_polygon_dggs_outer_linestring)
    disjoint_polygon_dggs_shape_polygon = DggsShape(
        disjoint_polygon_dggs_polygon, DggsShapeLocation.ONE_FACE
    )


## Unit tests for the Eaggr class.
class TestDggsPython(unittest.TestCase):
    _PROJ_PATH = os.path.join(os.path.dirname(__file__), "..", "..")

    def test_get_version(self):
        version = Eaggr(Model.ISEA4T).get_version()
        self.assertEqual(version, "v2.0")

    def test_convert_point_to_dggs_cell(self):
        # Create the lat/long points
        lat_long_point1 = LatLongPoint(1.234, 2.345, 3.884)
        lat_long_point2 = LatLongPoint(3.456, 4.567, 3.884)
        # Initialise the DGGS model
        dggs = Eaggr(Model.ISEA4T)
        # Convert the first lat/long point
        dggs_cell = dggs.convert_point_to_dggs_cell(lat_long_point1)
        self.assertTrue(type(dggs_cell) is DggsCell)
        self.assertEqual(dggs_cell.get_cell_id(), "07231131111113100331001")
        # Convert the second lat/long point
        dggs_cell = dggs.convert_point_to_dggs_cell(lat_long_point2)
        self.assertTrue(type(dggs_cell) is DggsCell)
        self.assertEqual(dggs_cell.get_cell_id(), "07012000001303022011321")

    def test_convert_shapes_to_dggs_shapes(self):
        # Create the lat/long shapes
        point1 = LatLongPoint(1.234, 2.345, 3.884)
        point2 = LatLongPoint(3.456, 4.567, 3.884)
        linestring = LatLongLinestring()
        linestring.add_point(point1)
        linestring.add_point(point2)
        polygon = LatLongPolygon(linestring)
        polygon.add_inner_ring(linestring)
        polygon.add_inner_ring(linestring)
        # Add the lat/long shapes to the input list
        lat_long_shapes = []
        lat_long_shapes.append(point1)
        lat_long_shapes.append(point2)
        lat_long_shapes.append(linestring)
        lat_long_shapes.append(polygon)
        # Convert the lat/long shapes
        dggs = Eaggr(Model.ISEA4T)
        dggs_shapes = dggs.convert_shapes_to_dggs_shapes(lat_long_shapes)
        # Check the number of shapes returned
        self.assertEqual(len(dggs_shapes), 4)
        # Check the first shape
        self.assertEqual(dggs_shapes[0].get_location(), DggsShapeLocation.ONE_FACE)
        self.assertEqual(dggs_shapes[0].get_shape_type(), DggsShapeType.CELL)
        self.assertTrue(type(dggs_shapes[0].get_shape()) is DggsCell)
        self.assertEqual(
            dggs_shapes[0].get_shape().get_cell_id(), "07231131111113100331001"
        )
        # Check the second shape
        self.assertEqual(dggs_shapes[1].get_location(), DggsShapeLocation.ONE_FACE)
        self.assertEqual(dggs_shapes[1].get_shape_type(), DggsShapeType.CELL)
        self.assertTrue(type(dggs_shapes[1].get_shape()) is DggsCell)
        self.assertEqual(
            dggs_shapes[1].get_shape().get_cell_id(), "07012000001303022011321"
        )
        # Check the third shape
        self.assertEqual(dggs_shapes[2].get_location(), DggsShapeLocation.NO_LOCATION)
        self.assertEqual(dggs_shapes[2].get_shape_type(), DggsShapeType.LINESTRING)
        self.assertTrue(type(dggs_shapes[2].get_shape()) is DggsLinestring)
        cells = dggs_shapes[2].get_shape().get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")
        # Check the fourth shape
        self.assertEqual(dggs_shapes[3].get_location(), DggsShapeLocation.NO_LOCATION)
        self.assertTrue(dggs_shapes[3].get_shape_type() is DggsShapeType.POLYGON)
        self.assertTrue(type(dggs_shapes[3].get_shape()) is DggsPolygon)
        cells = dggs_shapes[3].get_shape().get_outer_ring().get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")
        rings = dggs_shapes[3].get_shape().get_inner_rings()
        self.assertEqual(len(rings), 2)
        cells = rings[0].get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")
        cells = rings[1].get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")
        # Convert empty list
        dggs_shapes = dggs.convert_shapes_to_dggs_shapes([])
        self.assertEqual(len(dggs_shapes), 0)

    def test_convert_shape_string_to_dggs_shapes(self):
        wkt_string = (
            "GEOMETRYCOLLECTION("
            "MULTIPOINT(2.345 1.234, 4.567 3.456), "
            "LINESTRING(2.345 1.234, 4.567 3.456), "
            "POLYGON((2.345 1.234, 4.567 3.456), (2.345 1.234, 4.567 3.456), (2.345 1.234, 4.567 3.456)))"
        )
        # Convert the shape string
        dggs = Eaggr(Model.ISEA4T)
        dggs_shapes = dggs.convert_shape_string_to_dggs_shapes(
            wkt_string, ShapeStringFormat.WKT, 3.884
        )
        # Check the number of points returned
        self.assertEqual(len(dggs_shapes), 4)
        # Check the first shape
        self.assertEqual(dggs_shapes[0].get_location(), DggsShapeLocation.ONE_FACE)
        self.assertEqual(dggs_shapes[0].get_shape_type(), DggsShapeType.CELL)
        self.assertTrue(type(dggs_shapes[0].get_shape()) is DggsCell)
        self.assertEqual(
            dggs_shapes[0].get_shape().get_cell_id(), "07231131111113100331001"
        )
        # Check the second shape
        self.assertEqual(dggs_shapes[1].get_location(), DggsShapeLocation.ONE_FACE)
        self.assertEqual(dggs_shapes[1].get_shape_type(), DggsShapeType.CELL)
        self.assertTrue(type(dggs_shapes[1].get_shape()) is DggsCell)
        self.assertEqual(
            dggs_shapes[1].get_shape().get_cell_id(), "07012000001303022011321"
        )
        # Check the third shape
        self.assertEqual(dggs_shapes[2].get_location(), DggsShapeLocation.NO_LOCATION)
        self.assertEqual(dggs_shapes[2].get_shape_type(), DggsShapeType.LINESTRING)
        self.assertTrue(type(dggs_shapes[2].get_shape()) is DggsLinestring)
        cells = dggs_shapes[2].get_shape().get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")
        # Check the fourth shape
        self.assertEqual(dggs_shapes[3].get_location(), DggsShapeLocation.NO_LOCATION)
        self.assertEqual(dggs_shapes[3].get_shape_type(), DggsShapeType.POLYGON)
        self.assertTrue(type(dggs_shapes[3].get_shape()) is DggsPolygon)
        cells = dggs_shapes[3].get_shape().get_outer_ring().get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")
        rings = dggs_shapes[3].get_shape().get_inner_rings()
        self.assertEqual(len(rings), 2)
        cells = rings[0].get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")
        cells = rings[1].get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "07231131111113100331001")
        self.assertEqual(cells[1].get_cell_id(), "07012000001303022011321")

    def test_convert_dggs_cell_to_point(self):
        # Create the DGGS cells
        dggs_cell1 = DggsCell("07231131111113100331001")
        dggs_cell2 = DggsCell("07012000001303022011321")
        # Initialise the DGGS model
        dggs = Eaggr(Model.ISEA4T)
        # Convert the first DGGS cell
        lat_long_point = dggs.convert_dggs_cell_to_point(dggs_cell1)
        self.failUnlessAlmostEqual(1.234, lat_long_point.get_latitude(), places=3)
        self.failUnlessAlmostEqual(2.345, lat_long_point.get_longitude(), places=3)
        self.failUnlessAlmostEqual(5.804, lat_long_point.get_accuracy(), places=3)
        # Convert the second DGGS cell
        lat_long_point = dggs.convert_dggs_cell_to_point(dggs_cell2)
        self.failUnlessAlmostEqual(3.456, lat_long_point.get_latitude(), places=3)
        self.failUnlessAlmostEqual(4.567, lat_long_point.get_longitude(), places=3)
        self.failUnlessAlmostEqual(5.804, lat_long_point.get_accuracy(), places=3)

    def test_convert_dggs_cells_to_points(self):
        # Create the DGGS cells
        dggs_cells = [
            DggsCell("07231131111113100331001"),
            DggsCell("07012000001303022011321"),
        ]
        # Convert the DGGS cells
        dggs = Eaggr(Model.ISEA4T)
        lat_long_points = dggs.convert_dggs_cells_to_points(dggs_cells)
        # Check the number of points returned
        self.assertEqual(len(lat_long_points), 2)
        # Check the first point
        point_index = 0
        self.failUnlessAlmostEqual(
            1.234, lat_long_points[point_index].get_latitude(), places=3
        )
        self.failUnlessAlmostEqual(
            2.345, lat_long_points[point_index].get_longitude(), places=3
        )
        self.failUnlessAlmostEqual(
            5.804, lat_long_points[point_index].get_accuracy(), places=3
        )
        # Check the second point
        point_index += 1
        self.failUnlessAlmostEqual(
            3.456, lat_long_points[point_index].get_latitude(), places=3
        )
        self.failUnlessAlmostEqual(
            4.567, lat_long_points[point_index].get_longitude(), places=3
        )
        self.failUnlessAlmostEqual(
            5.804, lat_long_points[point_index].get_accuracy(), places=3
        )
        # Convert empty list
        lat_long_points = dggs.convert_dggs_cells_to_points([])
        self.assertEqual(len(lat_long_points), 0)

    def test_convert_dggs_cells_to_shape_string(self):
        # Create the DGGS cells
        dggs_cells = [
            DggsCell("07231131111113100331001"),
            DggsCell("07012000001303022011321"),
        ]
        # Convert the DGGS cells
        dggs = Eaggr(Model.ISEA4T)
        shape_string = dggs.convert_dggs_cells_to_shape_string(
            dggs_cells, ShapeStringFormat.WKT
        )
        # Split up the string and test the values numerically because values in the string can vary slightly
        string_data = re.split(" \(| |,|\)", shape_string)
        # Check the shape type
        self.assertEqual("MULTIPOINT", string_data[0])
        # Check the first point
        self.failUnlessAlmostEqual(2.345, float(string_data[1]), places=3)
        self.failUnlessAlmostEqual(1.234, float(string_data[2]), places=3)
        # Check the second point
        self.failUnlessAlmostEqual(4.567, float(string_data[3]), places=3)
        self.failUnlessAlmostEqual(3.456, float(string_data[4]), places=3)

    def test_convert_dggs_cell_outline_to_shape_string(self):
        # Create the DGGS cells
        dggs_cell = DggsCell("07231131111113100331001")
        # Convert the DGGS cells
        dggs = Eaggr(Model.ISEA4T)
        shape_string = dggs.convert_dggs_cell_outline_to_shape_string(
            dggs_cell, ShapeStringFormat.WKT
        )
        # Split up the string and test the values numerically because values in the string can vary slightly
        string_data = re.split(" \(\(| |,| |,|\)\)", shape_string)
        # Check the shape type
        self.assertEqual("POLYGON", string_data[0])
        # Check the first point
        self.failUnlessAlmostEqual(2.3449866, float(string_data[1]), places=7)
        self.failUnlessAlmostEqual(1.2340080, float(string_data[2]), places=7)
        # Check the second point
        self.failUnlessAlmostEqual(2.3450042, float(string_data[3]), places=7)
        self.failUnlessAlmostEqual(1.2339790, float(string_data[4]), places=7)
        # Check the second point
        self.failUnlessAlmostEqual(2.3450218, float(string_data[5]), places=7)
        self.failUnlessAlmostEqual(1.2340036, float(string_data[6]), places=7)

    def test_get_dggs_cell_parents(self):
        # Create the DGGS cell
        dggs_cell = DggsCell("07012212222221011101013")
        # Get its parent cell
        dggs = Eaggr(Model.ISEA4T)
        parents = dggs.get_dggs_cell_parents(dggs_cell)
        # Check the parent cell
        self.assertEqual(len(parents), 1)
        self.assertEqual(parents[0].get_cell_id(), "0701221222222101110101")

    def test_get_dggs_cell_children(self):
        # Create the DGGS cell
        dggs_cell = DggsCell("0701221222222101110101")
        # Get its child cells
        dggs = Eaggr(Model.ISEA4T)
        children = dggs.get_dggs_cell_children(dggs_cell)
        # Check the child cells
        self.assertEqual(len(children), 4)
        self.assertEqual(children[0].get_cell_id(), "07012212222221011101010")
        self.assertEqual(children[1].get_cell_id(), "07012212222221011101011")
        self.assertEqual(children[2].get_cell_id(), "07012212222221011101012")
        self.assertEqual(children[3].get_cell_id(), "07012212222221011101013")

    def test_get_dggs_cell_siblings(self):
        # Create the DGGS cell
        dggs_cell = DggsCell("07012212222221011101010")
        # Get its sibling cells
        dggs = Eaggr(Model.ISEA4T)
        siblings = dggs.get_dggs_cell_siblings(dggs_cell)
        # Check the sibling cells
        self.assertEqual(len(siblings), 3)
        self.assertEqual(siblings[0].get_cell_id(), "07012212222221011101011")
        self.assertEqual(siblings[1].get_cell_id(), "07012212222221011101012")
        self.assertEqual(siblings[2].get_cell_id(), "07012212222221011101013")

    def test_get_bounding_dggs_cell(self):
        # Create DGGS cells at the same resolution
        dggs_cells = [DggsCell("07001"), DggsCell("07002"), DggsCell("07003")]
        # Get their bounding cell
        dggs = Eaggr(Model.ISEA4T)
        bounding_cell = dggs.get_bounding_dggs_cell(dggs_cells)
        # Check the bounding cell
        self.assertEqual(bounding_cell.get_cell_id(), "0700")
        # Create DGGS cells at different resolutions
        dggs_cells = [
            DggsCell("0700123123"),
            DggsCell("0700231231231"),
            DggsCell("07003123123123123123123"),
        ]
        # Get their bounding cell
        bounding_cell = dggs.get_bounding_dggs_cell(dggs_cells)
        # Check the bounding cell
        self.assertEqual(bounding_cell.get_cell_id(), "0700")

    def test_create_dggs_kml_file(self):
        # Create the input DGGS cells
        dggs_cells = [
            DggsCell("07231131111113100331032"),
            DggsCell("07012000001303022011321"),
        ]
        # Get the directory path for the KML files
        test_data_dir = os.path.join(self._PROJ_PATH, "TestData")
        # Create a KML file for the DGGS cells
        dggs = Eaggr(Model.ISEA4T)
        kml_filename = os.path.join(test_data_dir, "ActualKmlFile.kml")
        dggs.create_dggs_kml_file(kml_filename, dggs_cells)
        # Check the contents of the output KML file
        with open(
            os.path.join(test_data_dir, "ExpectedDggsLibraryTestOutput.kml"), "r"
        ) as expected_file:
            expected_kml = expected_file.read()
        with open(kml_filename, "r") as actual_file:
            actual_kml = actual_file.read()
        os.remove(kml_filename)
        # Remove end of line characters from the strings
        expected_kml = expected_kml.replace("\r", "").replace("\n", "")
        actual_kml = actual_kml.replace("\r", "").replace("\n", "")
        self.assertEqual(expected_kml, actual_kml)

    def test_convert_shapes_to_dggs_shapes_isea3h(self):
        # Create the lat/long shapes
        point1 = LatLongPoint(1.234, 2.345, 3.884)
        point2 = LatLongPoint(3.456, 4.567, 3.884)
        linestring = LatLongLinestring()
        linestring.add_point(point1)
        linestring.add_point(point2)
        polygon = LatLongPolygon(linestring)
        polygon.add_inner_ring(linestring)
        polygon.add_inner_ring(linestring)
        # Add the lat/long shapes to the input list
        lat_long_shapes = []
        lat_long_shapes.append(point1)
        lat_long_shapes.append(point2)
        lat_long_shapes.append(linestring)
        lat_long_shapes.append(polygon)
        # Convert the lat/long shapes
        dggs = Eaggr(Model.ISEA3H)
        dggs_shapes = dggs.convert_shapes_to_dggs_shapes(lat_long_shapes)
        # Check the number of shapes returned
        self.assertEqual(len(dggs_shapes), 4)
        # Check the first shape
        self.assertEqual(dggs_shapes[0].get_location(), DggsShapeLocation.ONE_FACE)
        self.assertEqual(dggs_shapes[0].get_shape_type(), DggsShapeType.CELL)
        self.assertTrue(type(dggs_shapes[0].get_shape()) is DggsCell)
        self.assertEqual(
            dggs_shapes[0].get_shape().get_cell_id(), "0728-549628,-522499"
        )
        # Check the second shape
        self.assertEqual(dggs_shapes[1].get_location(), DggsShapeLocation.ONE_FACE)
        self.assertEqual(dggs_shapes[1].get_shape_type(), DggsShapeType.CELL)
        self.assertTrue(type(dggs_shapes[1].get_shape()) is DggsCell)
        self.assertEqual(
            dggs_shapes[1].get_shape().get_cell_id(), "0728-592991,-296234"
        )
        # Check the third shape
        self.assertEqual(dggs_shapes[2].get_location(), DggsShapeLocation.NO_LOCATION)
        self.assertEqual(dggs_shapes[2].get_shape_type(), DggsShapeType.LINESTRING)
        self.assertTrue(type(dggs_shapes[2].get_shape()) is DggsLinestring)
        cells = dggs_shapes[2].get_shape().get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "0728-549628,-522499")
        self.assertEqual(cells[1].get_cell_id(), "0728-592991,-296234")
        # Check the fourth shape
        self.assertEqual(dggs_shapes[3].get_location(), DggsShapeLocation.NO_LOCATION)
        self.assertTrue(dggs_shapes[3].get_shape_type() is DggsShapeType.POLYGON)
        self.assertTrue(type(dggs_shapes[3].get_shape()) is DggsPolygon)
        cells = dggs_shapes[3].get_shape().get_outer_ring().get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "0728-549628,-522499")
        self.assertEqual(cells[1].get_cell_id(), "0728-592991,-296234")
        rings = dggs_shapes[3].get_shape().get_inner_rings()
        self.assertEqual(len(rings), 2)
        cells = rings[0].get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "0728-549628,-522499")
        self.assertEqual(cells[1].get_cell_id(), "0728-592991,-296234")
        cells = rings[1].get_cells()
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].get_cell_id(), "0728-549628,-522499")
        self.assertEqual(cells[1].get_cell_id(), "0728-592991,-296234")

    ## Test to ensure multiple instances of the DGGS API can be used without conflicting
    #  with each other.
    def test_multiple_instances(self):
        # Create a point to convert
        lat_long_point = LatLongPoint(1.234, 2.345, 3.884)
        # Create two different DGGS models
        isea4tDggs = Eaggr(Model.ISEA4T)
        isea3hDggs = Eaggr(Model.ISEA3H)
        # Convert the point to a ISEA4T cell
        dggs_cell = isea4tDggs.convert_point_to_dggs_cell(lat_long_point)
        self.assertEqual(dggs_cell.get_cell_id(), "07231131111113100331001")
        # Convert the point to a ISEA3H cell
        dggs_cell = isea3hDggs.convert_point_to_dggs_cell(lat_long_point)
        self.assertEqual(dggs_cell.get_cell_id(), "0728-549628,-522499")
        # Now delete the ISEA4T model and ensure the ISEA3H model can still be used
        del isea4tDggs
        dggs_cell = isea3hDggs.convert_point_to_dggs_cell(lat_long_point)
        self.assertEqual(dggs_cell.get_cell_id(), "0728-549628,-522499")

    def test_compare_dggs_shapes_equals(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the EQUALS spatial analysis
        # Compare the cell with itself
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.dggs_shape_cell,
                DggsAnalysisType.EQUALS,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.EQUALS,
            )
        )
        # Compare the linestring with itself
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsAnalysisType.EQUALS,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.EQUALS,
            )
        )
        # Compare the polygon with itself
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsAnalysisType.EQUALS,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.EQUALS,
            )
        )

    def test_compare_dggs_shapes_contains(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the CONTAINS spatial analysis
        # Compare the cell with a child cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.child_dggs_shape_cell,
                DggsAnalysisType.CONTAINS,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.CONTAINS,
            )
        )
        # Compare the linestring with a substring
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.substring_dggs_shape_linestring,
                DggsAnalysisType.CONTAINS,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.CONTAINS,
            )
        )
        # Compare the polygon with a polygon in its interior
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.interior_polygon_dggs_shape_polygon,
                DggsAnalysisType.CONTAINS,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.CONTAINS,
            )
        )

    def test_compare_dggs_shapes_within(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the WITHIN spatial analysis
        # Compare the cell with a parent cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.parent_dggs_shape_cell,
                DggsAnalysisType.WITHIN,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.WITHIN,
            )
        )
        # Compare the linestring with a super-string
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.superstring_dggs_shape_linestring,
                DggsAnalysisType.WITHIN,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.WITHIN,
            )
        )
        # Compare the polygon with a polygon that contains it
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.contains_polygon_dggs_shape_polygon,
                DggsAnalysisType.WITHIN,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.WITHIN,
            )
        )

    def test_compare_dggs_shapes_touches(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the TOUCHES spatial analysis
        # Compare the cell with an edge sharing cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.edge_sharing_dggs_shape_cell,
                DggsAnalysisType.TOUCHES,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.TOUCHES,
            )
        )
        # Compare the linestring with a linestring that touches its end
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.ends_touching_dggs_shape_linestring,
                DggsAnalysisType.TOUCHES,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.TOUCHES,
            )
        )
        # Compare the polygon with a polygon that shares an edge of the outer ring
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.edge_sharing_polygon_dggs_shape_polygon,
                DggsAnalysisType.TOUCHES,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.TOUCHES,
            )
        )

    def test_compare_dggs_shapes_disjoint(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the DISJOINT spatial analysis
        # Compare the cell with a disjoint cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.DISJOINT,
            )
        )
        # Compare the cell with an edge sharing cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.edge_sharing_dggs_shape_cell,
                DggsAnalysisType.DISJOINT,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.DISJOINT,
            )
        )
        # Compare the linestring with a linestring that touches its end
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.ends_touching_dggs_shape_linestring,
                DggsAnalysisType.DISJOINT,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.DISJOINT,
            )
        )
        # Compare the polygon with a polygon that shares an edge of the outer ring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.edge_sharing_polygon_dggs_shape_polygon,
                DggsAnalysisType.DISJOINT,
            )
        )

    def test_compare_dggs_shapes_intersects(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the INTERSECTS spatial analysis
        # Compare the cell with an edge sharing cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.edge_sharing_dggs_shape_cell,
                DggsAnalysisType.INTERSECTS,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.INTERSECTS,
            )
        )
        # Compare the linestring with a linestring that touches its end
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.ends_touching_dggs_shape_linestring,
                DggsAnalysisType.INTERSECTS,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.INTERSECTS,
            )
        )
        # Compare the polygon with a polygon that shares an edge of the outer ring
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.edge_sharing_polygon_dggs_shape_polygon,
                DggsAnalysisType.INTERSECTS,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.INTERSECTS,
            )
        )

    def test_compare_dggs_shapes_covers(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the COVERS spatial analysis
        # Compare the cell with a child cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.child_dggs_shape_cell,
                DggsAnalysisType.COVERS,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.COVERS,
            )
        )
        # Compare the linestring with a substring
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.substring_dggs_shape_linestring,
                DggsAnalysisType.COVERS,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.COVERS,
            )
        )
        # Compare the polygon with a polygon in its interior
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.interior_polygon_dggs_shape_polygon,
                DggsAnalysisType.COVERS,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.COVERS,
            )
        )

    def test_compare_dggs_shapes_covered_by(self):
        dggs = Eaggr(Model.ISEA4T)
        # Test the COVERED_BY spatial analysis
        # Compare the cell with a parent cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.parent_dggs_shape_cell,
                DggsAnalysisType.COVERED_BY,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_cell,
                DggsTestDataTriangular.disjoint_dggs_shape_cell,
                DggsAnalysisType.COVERED_BY,
            )
        )
        # Compare the linestring with a super-string
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.superstring_dggs_shape_linestring,
                DggsAnalysisType.COVERED_BY,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_linestring,
                DggsTestDataTriangular.disjoint_dggs_shape_linestring,
                DggsAnalysisType.COVERED_BY,
            )
        )
        # Compare the polygon with a polygon that contains it
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.contains_polygon_dggs_shape_polygon,
                DggsAnalysisType.COVERED_BY,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataTriangular.dggs_shape_polygon,
                DggsTestDataTriangular.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.COVERED_BY,
            )
        )

    def test_compare_dggs_shapes_crosses(self):
        dggs = Eaggr(Model.ISEA3H)
        # Test the CROSSES spatial analysis
        # Two cells can never cross as they are treated as polygons
        # Compare the cell with an overlapping child cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_cell,
                DggsTestDataHexagonal.overlapping_child_dggs_shape_cell,
                DggsAnalysisType.CROSSES,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_cell,
                DggsTestDataHexagonal.disjoint_dggs_shape_cell,
                DggsAnalysisType.CROSSES,
            )
        )
        # Compare the linestring with a linestring that crosses it at a single point
        crossing_dggs_linestring = DggsLinestring()
        crossing_dggs_linestring.add_cell(DggsCell("07022,3"))
        crossing_dggs_linestring.add_cell(DggsCell("07023,1"))
        crossing_dggs_linestring.add_cell(DggsCell("07024,0"))
        crossing_dggs_shape_linestring = DggsShape(
            crossing_dggs_linestring, DggsShapeLocation.ONE_FACE
        )
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_linestring,
                crossing_dggs_shape_linestring,
                DggsAnalysisType.CROSSES,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_linestring,
                DggsTestDataHexagonal.disjoint_dggs_shape_linestring,
                DggsAnalysisType.CROSSES,
            )
        )
        # A polygon can never cross a polygon
        # Compare the polygon with a partially overlapping polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_polygon,
                DggsTestDataHexagonal.overlapping_polygon_dggs_shape_polygon,
                DggsAnalysisType.CROSSES,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_polygon,
                DggsTestDataHexagonal.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.CROSSES,
            )
        )

    def test_compare_dggs_shapes_overlaps(self):
        dggs = Eaggr(Model.ISEA3H)
        # Test the OVERLAPS spatial analysis
        # Compare the cell with an overlapping child cell
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_cell,
                DggsTestDataHexagonal.overlapping_child_dggs_shape_cell,
                DggsAnalysisType.OVERLAPS,
            )
        )
        # Compare the cell with a disjoint cell
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_cell,
                DggsTestDataHexagonal.disjoint_dggs_shape_cell,
                DggsAnalysisType.OVERLAPS,
            )
        )
        # Compare the linestring with a linestring that partially overlaps
        overlapping_dggs_linestring = DggsLinestring()
        overlapping_dggs_linestring.add_cell(DggsCell("07021,4"))
        overlapping_dggs_linestring.add_cell(DggsCell("07021,2"))
        overlapping_dggs_linestring.add_cell(DggsCell("07023,1"))
        overlapping_dggs_shape_linestring = DggsShape(
            overlapping_dggs_linestring, DggsShapeLocation.ONE_FACE
        )
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_linestring,
                overlapping_dggs_shape_linestring,
                DggsAnalysisType.OVERLAPS,
            )
        )
        # Compare the linestring with a disjoint linestring
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_linestring,
                DggsTestDataHexagonal.disjoint_dggs_shape_linestring,
                DggsAnalysisType.OVERLAPS,
            )
        )
        # Compare the polygon with a partially overlapping polygon
        self.assertTrue(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_polygon,
                DggsTestDataHexagonal.overlapping_polygon_dggs_shape_polygon,
                DggsAnalysisType.OVERLAPS,
            )
        )
        # Compare the polygon with a disjoint polygon
        self.assertFalse(
            dggs.compare_dggs_shapes(
                DggsTestDataHexagonal.dggs_shape_polygon,
                DggsTestDataHexagonal.disjoint_polygon_dggs_shape_polygon,
                DggsAnalysisType.OVERLAPS,
            )
        )



import sys
import argparse
import json
from tqdm import tqdm
from shapely.geometry import Polygon, shape, MultiPoint, MultiLineString, MultiPolygon
import pandas as pd
import geopandas as gpd
from vgrid.utils import geohash
from vgrid.generator.settings import graticule_dggs_to_feature
from vgrid.generator.geohashgrid import (
    initial_geohashes,
    geohash_to_polygon,
    expand_geohash_bbox,
)
from vgrid.conversion.dggscompact import geohashcompact
from vgrid.utils.geometry import (
    check_predicate,
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
)
from math import sqrt


def validate_geohash_resolution(resolution):
    """
    Validate that Geohash resolution is in the valid range [1..10] (1=coarsest, 10=finest).

    Args:
        resolution (int): Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [1..10]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution < 1 or resolution > 10:
        raise ValueError(f"Resolution must be in range [1..10], got {resolution}")

    return resolution


def point2geohash(
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,
):
    """
    Convert a point geometry to a Geohash grid cell.

    Args:
        resolution (int): Geohash resolution [1..10]
        point (shapely.geometry.Point): Point geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for points)
        compact (bool, optional): Enable Geohash compact mode (not used for points)
        topology (bool, optional): Enable topology preserving mode - ensures disjoint points have disjoint Geohash cells
        include_properties (bool, optional): Whether to include properties in output
        all_points: List of all points for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing Geohash cells containing the point
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        shortest_distance = shortest_point_distance(all_points)
        # Geohash cell size (approx): each character increases resolution, cell size shrinks by ~1/8 to 1/32
        # We'll use a rough estimate: cell size halves every character, so use a lookup or formula
        # For simplicity, use a fixed table for WGS84 (meters) for geohash length 1-10
        geohash_cell_sizes = [5000000, 1250000, 156000, 39100, 4890, 1220, 153, 38.2, 4.77, 1.19]  # meters
        for res in range(1, 11):
            cell_diameter = geohash_cell_sizes[res-1] * sqrt(2) * 2
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 10
    geohash_features = []
    longitude = point.x
    latitude = point.y
    geohash_id = geohash.encode(latitude, longitude, resolution)
    bbox = geohash.bbox(geohash_id)
    if bbox:
        min_lat, min_lon = bbox["s"], bbox["w"]
        max_lat, max_lon = bbox["n"], bbox["e"]
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],
                [max_lon, min_lat],
                [max_lon, max_lat],
                [min_lon, max_lat],
                [min_lon, min_lat],
            ]
        )
        geohash_feature = graticule_dggs_to_feature(
            "geohash", geohash_id, resolution, cell_polygon
        )
        if include_properties and feature_properties:
            geohash_feature["properties"].update(feature_properties)
        geohash_features.append(geohash_feature)
    return geohash_features


def polyline2geohash(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,
):
    """
    Convert line geometries (LineString, MultiLineString) to Geohash grid cells.

    Args:
        resolution (int): Geohash resolution [1..10]
        feature (shapely.geometry.LineString or shapely.geometry.MultiLineString): Line geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for lines)
        compact (bool, optional): Enable Geohash compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polylines have disjoint Geohash cells
        include_properties (bool, optional): Whether to include properties in output
        all_polylines: List of all polylines for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing Geohash cells intersecting the line
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        shortest_distance = shortest_polyline_distance(all_polylines)
        geohash_cell_sizes = [5000000, 1250000, 156000, 39100, 4890, 1220, 153, 38.2, 4.77, 1.19]  # meters
        for res in range(1, 11):
            cell_diameter = geohash_cell_sizes[res-1] * sqrt(2) * 4  # in case there are 2 cells representing the same line segment
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 10
    geohash_features = []
    if feature.geom_type in ("LineString"):
        polylines = [feature]
    elif feature.geom_type in ("MultiLineString"):
        polylines = list(feature.geoms)
    else:
        return []
    for polyline in polylines:
        intersected_geohashes = {
            gh
            for gh in initial_geohashes
            if geohash_to_polygon(gh).intersects(polyline)
        }
        geohashes_bbox = set()
        for gh in intersected_geohashes:
            expand_geohash_bbox(gh, resolution, geohashes_bbox, polyline)
        for gh in geohashes_bbox:
            cell_polygon = geohash_to_polygon(gh)
            geohash_feature = graticule_dggs_to_feature(
                "geohash", gh, resolution, cell_polygon
            )
            if include_properties and feature_properties:
                geohash_feature["properties"].update(feature_properties)
            geohash_features.append(geohash_feature)
    geohash_geojson = {"type": "FeatureCollection", "features": geohash_features}
    if compact:
        return geohashcompact(geohash_geojson)["features"]
    return geohash_features


def polygon2geohash(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,
):
    """
    Convert polygon geometries (Polygon, MultiPolygon) to Geohash grid cells.

    Args:
        resolution (int): Geohash resolution [1..10]
        feature (shapely.geometry.Polygon or shapely.geometry.MultiPolygon): Polygon geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply ('intersect', 'within', 'centroid_within', 'largest_overlap')
        compact (bool, optional): Enable Geohash compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polygons have disjoint Geohash cells
        include_properties (bool, optional): Whether to include properties in output
        all_polygons: List of all polygons for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing Geohash cells based on predicate
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        shortest_distance = shortest_polygon_distance(all_polygons)
        geohash_cell_sizes = [5000000, 1250000, 156000, 39100, 4890, 1220, 153, 38.2, 4.77, 1.19]  # meters
        for res in range(1, 11):
            cell_diameter = geohash_cell_sizes[res-1] * sqrt(2) * 4
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 10
    geohash_features = []
    if feature.geom_type in ("Polygon"):
        polygons = [feature]
    elif feature.geom_type in ("MultiPolygon"):
        polygons = list(feature.geoms)
    else:
        return []
    for polygon in polygons:
        intersected_geohashes = {
            gh for gh in initial_geohashes if geohash_to_polygon(gh).intersects(polygon)
        }
        geohashes_bbox = set()
        for gh in intersected_geohashes:
            expand_geohash_bbox(gh, resolution, geohashes_bbox, polygon)
        for gh in geohashes_bbox:
            cell_polygon = geohash_to_polygon(gh)
            geohash_feature = graticule_dggs_to_feature(
                "geohash", gh, resolution, cell_polygon
            )
            cell_geom = shape(geohash_feature["geometry"])
            if not check_predicate(cell_geom, polygon, predicate):
                continue
            if include_properties and feature_properties:
                geohash_feature["properties"].update(feature_properties)
            geohash_features.append(geohash_feature)
    geohash_geojson = {"type": "FeatureCollection", "features": geohash_features}
    if compact:
        return geohashcompact(geohash_geojson)["features"]
    return geohash_features


# --- Main geometry conversion ---
def geometry2geohash(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a list of geometries to Geohash grid cells.

    Args:
        geometries (shapely.geometry.BaseGeometry or list): Single geometry or list of geometries
        resolution (int): Geohash resolution [1..10]
        properties_list (list, optional): List of property dictionaries for each geometry
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable Geohash compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with Geohash grid cells
    """
    resolution = validate_geohash_resolution(resolution)
    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    geohash_features = []
    for i, geom in enumerate(tqdm(geometries, desc="Processing features")):
        if geom is None:
            continue
        props = properties_list[i] if i < len(properties_list) else {}
        if not include_properties:
            props = {}
        if geom.geom_type == "Point":
            geohash_features.extend(
                point2geohash(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,  # Pass all points for topology preservation
                )
            )
        elif geom.geom_type == "MultiPoint":
            for pt in geom.geoms:
                geohash_features.extend(
                    point2geohash(
                        resolution,
                        pt,
                        props,
                        predicate,
                        compact,
                        topology,
                        include_properties,
                        all_points,  # Pass all points for topology preservation
                    )
                )
        elif geom.geom_type in ("LineString", "MultiLineString"):
            geohash_features.extend(
                polyline2geohash(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polylines,  # Pass all polylines for topology preservation
                )
            )
        elif geom.geom_type in ("Polygon", "MultiPolygon"):
            geohash_features.extend(
                polygon2geohash(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polygons=all_polygons,  # Pass all polygons for topology preservation
                )
            )
    return {"type": "FeatureCollection", "features": geohash_features}


# --- DataFrame/GeoDataFrame conversion ---
def dataframe2geohash(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a pandas DataFrame with geometry column to Geohash grid cells.

    Args:
        df (pandas.DataFrame): DataFrame with geometry column
        resolution (int): Geohash resolution [1..10]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable Geohash compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with Geohash grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in df.iterrows():
        geom = row.geometry if "geometry" in row else row["geometry"]
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2geohash(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2geohash(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a GeoDataFrame to Geohash grid cells.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame to convert
        resolution (int): Geohash resolution [1..10]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable Geohash compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with Geohash grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in gdf.iterrows():
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2geohash(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


# --- Main vector2geohash function ---
def vector2geohash(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    """
    Convert vector data to Geohash grid cells from various input formats.

    Args:
        data: File path, URL, DataFrame, GeoJSON dict, or Shapely geometry
        resolution (int): Geohash resolution [1..10]
        compact (bool): Enable Geohash compact mode for polygons (default: False)
        output_format (str): Output format ('geojson', 'gpkg', 'parquet', 'csv', 'shapefile')
        output_path (str): Output file path (optional)
        include_properties (bool): If False, do not include original feature properties. (default: True)
        **kwargs: Additional arguments passed to geopandas read functions
    Returns:
        dict or str: Output in the specified format
    """

    # Process input data directly
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        # GeoDataFrame
        result = geodataframe2geohash(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        # DataFrame with geometry column
        result = dataframe2geohash(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        # Shapely geometry objects
        result = geometry2geohash(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        # GeoJSON dict
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2geohash(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    elif isinstance(data, str):
        # File path or URL
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2geohash(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")
    return convert_to_output_format(result, output_format, output_path)


# --- Output format conversion ---
def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert Geohash result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since Geohash uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2geohash.gpkg", driver="GPKG")
            return "vector2geohash.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2geohash.parquet", index=False)
            return "vector2geohash.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2geohash.shp", driver="ESRI Shapefile")
            return "vector2geohash.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


# --- CLI ---
def vector2geohash_cli():
    """
    Command-line interface for vector2geohash conversion.

    Usage:
        python vector2geohash.py -i input.shp -r 5 -c -f geojson -o output.geojson

    Arguments:
        -i, --input: Input file path or URL
        -r, --resolution: Geohash resolution [1..10]
        -c, --compact: Enable Geohash compact mode
        -p, --predicate: Spatial predicate (intersect, within, centroid_within, largest_overlap)
        -t, --topology: Enable topology preserving mode
        -np, --no-props: Do not include original feature properties
        -f, --format: Output format (geojson, gpkg, parquet, csv, shapefile)
        -o, --output: Output file path
    """
    parser = argparse.ArgumentParser(
        description="Convert vector data to Geohash grid cells"
    )
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=range(1, 11),
        metavar="[1-10]",
        help="Geohash resolution [1..10] (1=coarsest, 10=finest)",
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable Geohash compact mode for polygons",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", help="Enable topology preserving mode"
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")
    args = parser.parse_args()
    if args.resolution is not None:
        try:
            args.resolution = validate_geohash_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

    data = args.input
    output_path = args.output
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2geohash{extensions.get(args.format, '')}"
    try:
        vector2geohash(
            data,
            args.resolution,
            predicate=args.predicate,
            compact=args.compact,
            topology=args.topology,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )
        if output_path:
            print(f"Output saved to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2geohash_cli()



# Reference:
# https://github.com/aaliddell/s2cell,
# https://medium.com/@claude.ducharme/selecting-a-geo-representation-81afeaf3bf01
# https://github.com/sidewalklabs/s2
# https://github.com/google/s2geometry/tree/master/src/python
# https://github.com/google/s2geometry
# https://gis.stackexchange.com/questions/293716/creating-shapefile-of-s2-cells-for-given-level
# https://s2.readthedocs.io/en/latest/quickstart.html
from vgrid.utils import s2
import json
from tqdm import tqdm
from shapely.geometry import Polygon, mapping


def cell_to_polygon(cell_id):
    cell = s2.Cell(cell_id)
    vertices = []
    for i in range(4):
        vertex = s2.LatLng.from_point(cell.get_vertex(i))
        vertices.append((vertex.lng().degrees, vertex.lat().degrees))

    vertices.append(vertices[0])  # Close the polygon

    # Create a Shapely Polygon
    polygon = Polygon(vertices)
    return polygon
    # #  Fix Antimerididan:
    # fixed_polygon = fix_polygon(polygon)
    # return fixed_polygon


def generate_grid():
    # Define the cell level (S2 uses a level system for zoom, where level 30 is the highest resolution)
    level = 0
    # Create a list to store the S2 cell IDs
    cell_ids = []

    # Define the cell covering
    coverer = s2.RegionCoverer()
    coverer.min_level = level
    coverer.max_level = level
    # coverer.max_cells = 1_000_000  # Adjust as needed

    # Define the region to cover (in this example, we'll use the entire world)
    region = s2.LatLngRect(
        s2.LatLng.from_degrees(-90, -180), s2.LatLng.from_degrees(90, 180)
    )

    # Get the covering cells
    covering = coverer.get_covering(region)

    # Convert the covering cells to S2 cell IDs
    for cell_id in covering:
        cell_ids.append(cell_id)

    features = []
    for cell_id in tqdm(cell_ids, desc="Processing cells"):
        # Generate a Shapely Polygon
        polygon = cell_to_polygon(cell_id)

        # Convert Shapely Polygon to GeoJSON-like format using mapping()
        geometry = mapping(polygon)

        # Create a feature dictionary
        feature = {
            "type": "Feature",
            "geometry": geometry,
            "properties": {"s2": cell_id.to_token()},
        }

        features.append(feature)

    # Create a FeatureCollection
    return {"type": "FeatureCollection", "features": features}


def main():
    geojson_features = generate_grid()
    # Define the GeoJSON file path
    geojson_path = "cube.geojson"
    with open(geojson_path, "w") as f:
        json.dump(geojson_features, f, indent=2)

    print(f"Cube saved as {geojson_path}")


if __name__ == "__main__":
    main()



#  -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
#
#
# Convert locations to and from short codes.
#
# Open Location Codes are short, 10-11 character codes that can be used instead
# of street addresses. The codes can be generated and decoded offline, and use
# a reduced character set that minimises the chance of codes including words.
#
# Codes are able to be shortened relative to a nearby location. This means that
# in many cases, only four to seven characters of the code are needed.
# To recover the original code, the same location is not required, as long as
# a nearby location is provided.
#
# Codes represent rectangular areas rather than points, and the longer the
# code, the smaller the area. A 10 character code represents a 13.5x13.5
# meter area (at the equator. An 11 character code represents approximately
# a 2.8x3.5 meter area.
#
# Two encoding algorithms are used. The first 10 characters are pairs of
# characters, one for latitude and one for longitude, using base 20. Each pair
# reduces the area of the code by a factor of 400. Only even code lengths are
# sensible, since an odd-numbered length would have sides in a ratio of 20:1.
#
# At position 11, the algorithm changes so that each character selects one
# position from a 4x5 grid. This allows single-character refinements.
#
# Examples:
#
#   Encode a location, default accuracy:
#   encode(47.365590, 8.524997)
#
#   Encode a location using one stage of additional refinement:
#   encode(47.365590, 8.524997, 11)
#
#   Decode a full code:
#   coord = decode(code)
#   msg = "Center is {lat}, {lon}".format(lat=coord.latitudeCenter, lon=coord.longitudeCenter)
#
#   Attempt to trim the first characters from a code:
#   shorten('8FVC9G8F+6X', 47.5, 8.5)
#
#   Recover the full code from a short code:
#   recoverNearest('9G8F+6X', 47.4, 8.6)
#   recoverNearest('8F+6X', 47.4, 8.6)

import re
import math

# A separator used to break the code into two parts to aid memorability.
SEPARATOR_ = "+"

# The number of characters to place before the separator.
SEPARATOR_POSITION_ = 8

# The character used to pad codes.
PADDING_CHARACTER_ = "0"

# The character set used to encode the values.
CODE_ALPHABET_ = "23456789CFGHJMPQRVWX"

# The base to use to convert numbers to/from.
ENCODING_BASE_ = len(CODE_ALPHABET_)

# The maximum value for latitude in degrees.
LATITUDE_MAX_ = 90

# The maximum value for longitude in degrees.
LONGITUDE_MAX_ = 180

# The max number of digits to process in a plus code.
MAX_DIGIT_COUNT_ = 15

# Maximum code length using lat/lng pair encoding. The area of such a
# code is approximately 13x13 meters (at the equator), and should be suitable
# for identifying buildings. This excludes prefix and separator characters.
PAIR_CODE_LENGTH_ = 10

# First place value of the pairs (if the last pair value is 1).
PAIR_FIRST_PLACE_VALUE_ = ENCODING_BASE_ ** (PAIR_CODE_LENGTH_ / 2 - 1)

# Inverse of the precision of the pair section of the code.
PAIR_PRECISION_ = ENCODING_BASE_**3

# The resolution values in degrees for each position in the lat/lng pair
# encoding. These give the place value of each position, and therefore the
# dimensions of the resulting area.
PAIR_RESOLUTIONS_ = [20.0, 1.0, 0.05, 0.0025, 0.000125]

# Number of digits in the grid precision part of the code.
GRID_CODE_LENGTH_ = MAX_DIGIT_COUNT_ - PAIR_CODE_LENGTH_

# Number of columns in the grid refinement method.
GRID_COLUMNS_ = 4

# Number of rows in the grid refinement method.
GRID_ROWS_ = 5

# First place value of the latitude grid (if the last place is 1).
GRID_LAT_FIRST_PLACE_VALUE_ = GRID_ROWS_ ** (GRID_CODE_LENGTH_ - 1)

# First place value of the longitude grid (if the last place is 1).
GRID_LNG_FIRST_PLACE_VALUE_ = GRID_COLUMNS_ ** (GRID_CODE_LENGTH_ - 1)

# Multiply latitude by this much to make it a multiple of the finest
# precision.
FINAL_LAT_PRECISION_ = PAIR_PRECISION_ * GRID_ROWS_ ** (
    MAX_DIGIT_COUNT_ - PAIR_CODE_LENGTH_
)

# Multiply longitude by this much to make it a multiple of the finest
# precision.
FINAL_LNG_PRECISION_ = PAIR_PRECISION_ * GRID_COLUMNS_ ** (
    MAX_DIGIT_COUNT_ - PAIR_CODE_LENGTH_
)

# Minimum length of a code that can be shortened.
MIN_TRIMMABLE_CODE_LEN_ = 6

GRID_SIZE_DEGREES_ = 0.000125


def isValid(code):
    """
    Determines if a code is valid.
    To be valid, all characters must be from the Open Location Code character
    set with at most one separator. The separator can be in any even-numbered
    position up to the eighth digit.
    """
    # The separator is required.
    sep = code.find(SEPARATOR_)
    if code.count(SEPARATOR_) > 1:
        return False
    # Is it the only character?
    if len(code) == 1:
        return False
    # Is it in an illegal position?
    if sep == -1 or sep > SEPARATOR_POSITION_ or sep % 2 == 1:
        return False
    # We can have an even number of padding characters before the separator,
    # but then it must be the final character.
    pad = code.find(PADDING_CHARACTER_)
    if pad != -1:
        # Short codes cannot have padding
        if sep < SEPARATOR_POSITION_:
            return False
        # Not allowed to start with them!
        if pad == 0:
            return False

        # There can only be one group and it must have even length.
        rpad = code.rfind(PADDING_CHARACTER_) + 1
        pads = code[pad:rpad]
        if len(pads) % 2 == 1 or pads.count(PADDING_CHARACTER_) != len(pads):
            return False
        # If the code is long enough to end with a separator, make sure it does.
        if not code.endswith(SEPARATOR_):
            return False
    # If there are characters after the separator, make sure there isn't just
    # one of them (not legal).
    if len(code) - sep - 1 == 1:
        return False
    # Check the code contains only valid characters.
    sepPad = SEPARATOR_ + PADDING_CHARACTER_
    for ch in code:
        if ch.upper() not in CODE_ALPHABET_ and ch not in sepPad:
            return False
    return True


def isShort(code):
    """
    Determines if a code is a valid short code.
    A short Open Location Code is a sequence created by removing four or more
    digits from an Open Location Code. It must include a separator
    character.
    """
    # Check it's valid.
    if not isValid(code):
        return False
    # If there are less characters than expected before the SEPARATOR.
    sep = code.find(SEPARATOR_)
    if sep >= 0 and sep < SEPARATOR_POSITION_:
        return True
    return False


def isFull(code):
    """
    Determines if a code is a valid full Open Location Code.
    Not all possible combinations of Open Location Code characters decode to
    valid latitude and longitude values. This checks that a code is valid
    and also that the latitude and longitude values are legal. If the prefix
    character is present, it must be the first character. If the separator
    character is present, it must be after four characters.
    """
    if not isValid(code):
        return False
    # If it's short, it's not full
    if isShort(code):
        return False
    # Work out what the first latitude character indicates for latitude.
    firstLatValue = CODE_ALPHABET_.find(code[0].upper()) * ENCODING_BASE_
    if firstLatValue >= LATITUDE_MAX_ * 2:
        # The code would decode to a latitude of >= 90 degrees.
        return False
    if len(code) > 1:
        # Work out what the first longitude character indicates for longitude.
        firstLngValue = CODE_ALPHABET_.find(code[1].upper()) * ENCODING_BASE_
    if firstLngValue >= LONGITUDE_MAX_ * 2:
        # The code would decode to a longitude of >= 180 degrees.
        return False
    return True


def encode(latitude, longitude, codeLength=PAIR_CODE_LENGTH_):
    """
    Encode a location into an Open Location Code.
    Produces a code of the specified length, or the default length if no length
    is provided.
    The length determines the accuracy of the code. The default length is
    10 characters, returning a code of approximately 13.5x13.5 meters. Longer
    codes represent smaller areas, but lengths > 14 are sub-centimetre and so
    11 or 12 are probably the limit of useful codes.
    Args:
      latitude: A latitude in signed decimal degrees. Will be clipped to the
          range -90 to 90.
      longitude: A longitude in signed decimal degrees. Will be normalised to
          the range -180 to 180.
      codeLength: The number of significant digits in the output code, not
          including any separator characters.
    """
    if codeLength < 2 or (codeLength < PAIR_CODE_LENGTH_ and codeLength % 2 == 1):
        raise ValueError("Invalid Open Location Code length - " + str(codeLength))
    codeLength = min(codeLength, MAX_DIGIT_COUNT_)
    # Ensure that latitude and longitude are valid.
    latitude = clipLatitude(latitude)
    longitude = normalizeLongitude(longitude)
    # Latitude 90 needs to be adjusted to be just less, so the returned code
    # can also be decoded.
    if latitude == 90:
        latitude = latitude - computeLatitudePrecision(codeLength)
    code = ""

    # Compute the code.
    # This approach converts each value to an integer after multiplying it by
    # the final precision. This allows us to use only integer operations, so
    # avoiding any accumulation of floating point representation errors.

    # Multiply values by their precision and convert to positive.
    # Force to integers so the division operations will have integer results.
    # Note: Python requires rounding before truncating to ensure precision!
    latVal = int(round((latitude + LATITUDE_MAX_) * FINAL_LAT_PRECISION_, 6))
    lngVal = int(round((longitude + LONGITUDE_MAX_) * FINAL_LNG_PRECISION_, 6))

    # Compute the grid part of the code if necessary.
    if codeLength > PAIR_CODE_LENGTH_:
        for i in range(0, MAX_DIGIT_COUNT_ - PAIR_CODE_LENGTH_):
            latDigit = latVal % GRID_ROWS_
            lngDigit = lngVal % GRID_COLUMNS_
            ndx = latDigit * GRID_COLUMNS_ + lngDigit
            code = CODE_ALPHABET_[ndx] + code
            latVal //= GRID_ROWS_
            lngVal //= GRID_COLUMNS_
    else:
        latVal //= pow(GRID_ROWS_, GRID_CODE_LENGTH_)
        lngVal //= pow(GRID_COLUMNS_, GRID_CODE_LENGTH_)
    # Compute the pair section of the code.
    for i in range(0, PAIR_CODE_LENGTH_ // 2):
        code = CODE_ALPHABET_[lngVal % ENCODING_BASE_] + code
        code = CODE_ALPHABET_[latVal % ENCODING_BASE_] + code
        latVal //= ENCODING_BASE_
        lngVal //= ENCODING_BASE_

    # Add the separator character.
    code = code[:SEPARATOR_POSITION_] + SEPARATOR_ + code[SEPARATOR_POSITION_:]

    # If we don't need to pad the code, return the requested section.
    if codeLength >= SEPARATOR_POSITION_:
        return code[0 : codeLength + 1]

    # Pad and return the code.
    return code[0:codeLength] + "".zfill(SEPARATOR_POSITION_ - codeLength) + SEPARATOR_


def decode(code):
    """
    Decodes an Open Location Code into the location coordinates.
    Returns a CodeArea object that includes the coordinates of the bounding
    box - the lower left, center and upper right.
    Args:
      code: The Open Location Code to decode.
    Returns:
      A CodeArea object that provides the latitude and longitude of two of the
      corners of the area, the center, and the length of the original code.
    """
    if not isFull(code):
        raise ValueError(
            "Passed Open Location Code is not a valid full code - " + str(code)
        )

    # Strip out separator character (we've already established the code is
    # valid so the maximum is one), and padding characters. Convert to upper
    # case and constrain to the maximum number of digits.
    code = re.sub("[+0]", "", code)
    code = code.upper()
    code = code[:MAX_DIGIT_COUNT_]
    # Initialise the values for each section. We work them out as integers and
    # convert them to floats at the end.
    normalLat = -LATITUDE_MAX_ * PAIR_PRECISION_
    normalLng = -LONGITUDE_MAX_ * PAIR_PRECISION_
    gridLat = 0
    gridLng = 0
    # How many digits do we have to process?
    digits = min(len(code), PAIR_CODE_LENGTH_)
    # Define the place value for the most significant pair.
    pv = PAIR_FIRST_PLACE_VALUE_
    # Decode the paired digits.
    for i in range(0, digits, 2):
        normalLat += CODE_ALPHABET_.find(code[i]) * pv
        normalLng += CODE_ALPHABET_.find(code[i + 1]) * pv
        if i < digits - 2:
            pv //= ENCODING_BASE_

    # Convert the place value to a float in degrees.
    latPrecision = float(pv) / PAIR_PRECISION_
    lngPrecision = float(pv) / PAIR_PRECISION_
    # Process any extra precision digits.
    if len(code) > PAIR_CODE_LENGTH_:
        # Initialise the place values for the grid.
        rowpv = GRID_LAT_FIRST_PLACE_VALUE_
        colpv = GRID_LNG_FIRST_PLACE_VALUE_
        # How many digits do we have to process?
        digits = min(len(code), MAX_DIGIT_COUNT_)
        for i in range(PAIR_CODE_LENGTH_, digits):
            digitVal = CODE_ALPHABET_.find(code[i])
            row = digitVal // GRID_COLUMNS_
            col = digitVal % GRID_COLUMNS_
            gridLat += row * rowpv
            gridLng += col * colpv
            if i < digits - 1:
                rowpv //= GRID_ROWS_
                colpv //= GRID_COLUMNS_

        # Adjust the precisions from the integer values to degrees.
        latPrecision = float(rowpv) / FINAL_LAT_PRECISION_
        lngPrecision = float(colpv) / FINAL_LNG_PRECISION_

    # Merge the values from the normal and extra precision parts of the code.
    lat = float(normalLat) / PAIR_PRECISION_ + float(gridLat) / FINAL_LAT_PRECISION_
    lng = float(normalLng) / PAIR_PRECISION_ + float(gridLng) / FINAL_LNG_PRECISION_
    # Multiple values by 1e14, round and then divide. This reduces errors due
    # to floating point precision.
    return CodeArea(
        round(lat, 14),
        round(lng, 14),
        round(lat + latPrecision, 14),
        round(lng + lngPrecision, 14),
        min(len(code), MAX_DIGIT_COUNT_),
    )


def recoverNearest(code, referenceLatitude, referenceLongitude):
    """
    Recover the nearest matching code to a specified location.
    Given a short code of between four and seven characters, this recovers
    the nearest matching full code to the specified location.
    Args:
      code: A valid OLC character sequence.
      referenceLatitude: The latitude (in signed decimal degrees) to use to
          find the nearest matching full code.
      referenceLongitude: The longitude (in signed decimal degrees) to use
          to find the nearest matching full code.
    Returns:
      The nearest full Open Location Code to the reference location that matches
      the short code. If the passed code was not a valid short code, but was a
      valid full code, it is returned with proper capitalization but otherwise
      unchanged.
    """
    # if code is a valid full code, return it properly capitalized
    if isFull(code):
        return code.upper()
    if not isShort(code):
        raise ValueError("Passed short code is not valid - " + str(code))
    # Ensure that latitude and longitude are valid.
    referenceLatitude = clipLatitude(referenceLatitude)
    referenceLongitude = normalizeLongitude(referenceLongitude)
    # Clean up the passed code.
    code = code.upper()
    # Compute the number of digits we need to recover.
    paddingLength = SEPARATOR_POSITION_ - code.find(SEPARATOR_)
    # The resolution (height and width) of the padded area in degrees.
    resolution = pow(20, 2 - (paddingLength / 2))
    # Distance from the center to an edge (in degrees).
    halfResolution = resolution / 2.0
    # Use the reference location to pad the supplied short code and decode it.
    codeArea = decode(
        encode(referenceLatitude, referenceLongitude)[0:paddingLength] + code
    )
    # How many degrees latitude is the code from the reference? If it is more
    # than half the resolution, we need to move it north or south but keep it
    # within -90 to 90 degrees.
    if (
        referenceLatitude + halfResolution < codeArea.latitudeCenter
        and codeArea.latitudeCenter - resolution >= -LATITUDE_MAX_
    ):
        # If the proposed code is more than half a cell north of the reference location,
        # it's too far, and the best match will be one cell south.
        codeArea.latitudeCenter -= resolution
    elif (
        referenceLatitude - halfResolution > codeArea.latitudeCenter
        and codeArea.latitudeCenter + resolution <= LATITUDE_MAX_
    ):
        # If the proposed code is more than half a cell south of the reference location,
        # it's too far, and the best match will be one cell north.
        codeArea.latitudeCenter += resolution
    # Adjust longitude if necessary.
    if referenceLongitude + halfResolution < codeArea.longitudeCenter:
        codeArea.longitudeCenter -= resolution
    elif referenceLongitude - halfResolution > codeArea.longitudeCenter:
        codeArea.longitudeCenter += resolution
    return encode(
        codeArea.latitudeCenter, codeArea.longitudeCenter, codeArea.codeLength
    )


def shorten(code, latitude, longitude):
    """
    Remove characters from the start of an OLC code.
    This uses a reference location to determine how many initial characters
    can be removed from the OLC code. The number of characters that can be
    removed depends on the distance between the code center and the reference
    location.
    The minimum number of characters that will be removed is four. If more than
    four characters can be removed, the additional characters will be replaced
    with the padding character. At most eight characters will be removed.
    The reference location must be within 50% of the maximum range. This ensures
    that the shortened code will be able to be recovered using slightly different
    locations.
    Args:
      code: A full, valid code to shorten.
      latitude: A latitude, in signed decimal degrees, to use as the reference
          point.
      longitude: A longitude, in signed decimal degrees, to use as the reference
          point.
    Returns:
      Either the original code, if the reference location was not close enough,
      or the .
    """
    if not isFull(code):
        raise ValueError("Passed code is not valid and full: " + str(code))
    if code.find(PADDING_CHARACTER_) != -1:
        raise ValueError("Cannot shorten padded codes: " + str(code))
    code = code.upper()
    codeArea = decode(code)
    if codeArea.codeLength < MIN_TRIMMABLE_CODE_LEN_:
        raise ValueError("Code length must be at least " + MIN_TRIMMABLE_CODE_LEN_)
    # Ensure that latitude and longitude are valid.
    latitude = clipLatitude(latitude)
    longitude = normalizeLongitude(longitude)
    # How close are the latitude and longitude to the code center.
    coderange = max(
        abs(codeArea.latitudeCenter - latitude),
        abs(codeArea.longitudeCenter - longitude),
    )
    for i in range(len(PAIR_RESOLUTIONS_) - 2, 0, -1):
        # Check if we're close enough to shorten. The range must be less than 1/2
        # the resolution to shorten at all, and we want to allow some safety, so
        # use 0.3 instead of 0.5 as a multiplier.
        if coderange < (PAIR_RESOLUTIONS_[i] * 0.3):
            # Trim it.
            return code[(i + 1) * 2 :]
    return code


def clipLatitude(latitude):
    """
    Clip a latitude into the range -90 to 90.
    Args:
      latitude: A latitude in signed decimal degrees.
    """
    return min(90, max(-90, latitude))


def computeLatitudePrecision(codeLength):
    """
    Compute the latitude precision value for a given code length. Lengths <=
    10 have the same precision for latitude and longitude, but lengths > 10
    have different precisions due to the grid method having fewer columns than
    rows.
    """
    if codeLength <= 10:
        return pow(20, math.floor((codeLength / -2) + 2))
    return pow(20, -3) / pow(GRID_ROWS_, codeLength - 10)


def normalizeLongitude(longitude):
    """
    Normalize a longitude into the range -180 to 180, not including 180.
    Args:
      longitude: A longitude in signed decimal degrees.
    """
    while longitude < -180:
        longitude = longitude + 360
    while longitude >= 180:
        longitude = longitude - 360
    return longitude


class CodeArea(object):
    """
    Coordinates of a decoded Open Location Code.
    The coordinates include the latitude and longitude of the lower left and
    upper right corners and the center of the bounding box for the area the
    code represents.
    Attributes:
      latitude_lo: The latitude of the SW corner in degrees.
      longitude_lo: The longitude of the SW corner in degrees.
      latitude_hi: The latitude of the NE corner in degrees.
      longitude_hi: The longitude of the NE corner in degrees.
      latitude_center: The latitude of the center in degrees.
      longitude_center: The longitude of the center in degrees.
      code_length: The number of significant characters that were in the code.
          This excludes the separator.
    """

    def __init__(self, latitudeLo, longitudeLo, latitudeHi, longitudeHi, codeLength):
        self.latitudeLo = latitudeLo
        self.longitudeLo = longitudeLo
        self.latitudeHi = latitudeHi
        self.longitudeHi = longitudeHi
        self.codeLength = codeLength
        self.latitudeCenter = min(
            latitudeLo + (latitudeHi - latitudeLo) / 2, LATITUDE_MAX_
        )
        self.longitudeCenter = min(
            longitudeLo + (longitudeHi - longitudeLo) / 2, LONGITUDE_MAX_
        )

    def __repr__(self):
        return str(
            [
                self.latitudeLo,
                self.longitudeLo,
                self.latitudeHi,
                self.longitudeHi,
                self.latitudeCenter,
                self.longitudeCenter,
                self.codeLength,
            ]
        )

    def latlng(self):
        return [self.latitudeCenter, self.longitudeCenter]


################
# Added by Vgrid
################


def olc_to_latlon(olc_id):
    code_area = decode(olc_id)
    center_lat = (code_area.latitudeLo + code_area.latitudeHi) / 2
    center_lng = (code_area.longitudeLo + code_area.longitudeHi) / 2
    return center_lat, center_lng


def olc_parent(olc_id):
    if not isFull(olc_id):
        raise ValueError(
            f"Passed Open Location Code is not a valid full code - {olc_id}"
        )

    code_area = decode(olc_id)
    current_len = code_area.codeLength

    if current_len <= 2:
        raise ValueError("No parent exists for a 2-digit OLC")

    # Determine parent resolution
    if current_len <= 10:
        parent_len = current_len - 2
    else:
        parent_len = current_len - 1

    # Encode center at parent precision
    lat_center = (code_area.latitudeLo + code_area.latitudeHi) / 2
    lon_center = (code_area.longitudeLo + code_area.longitudeHi) / 2
    parent_code = encode(lat_center, lon_center, parent_len)

    return parent_code


def olc_children(olc_id, resolution):
    if not isFull(olc_id):
        raise ValueError("Not a valid full OLC")

    if resolution not in [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]:
        raise ValueError("Resolution must be in [2, 4, 6, 8, 10...15]")

    parent = decode(olc_id)
    lat_center = (parent.latitudeLo + parent.latitudeHi) / 2
    lon_center = (parent.longitudeLo + parent.longitudeHi) / 2

    parent_code_len = parent.codeLength
    if resolution <= parent_code_len:
        raise ValueError("Resolution must be greater than current OLC precision")

    # Get dimensions of child at target resolution
    res_box = decode(encode(lat_center, lon_center, resolution))
    lat_step = res_box.latitudeHi - res_box.latitudeLo
    lon_step = res_box.longitudeHi - res_box.longitudeLo

    # Determine how many steps in each direction (for simplicity, full coverage)
    lat_count = round((parent.latitudeHi - parent.latitudeLo) / lat_step)
    lon_count = round((parent.longitudeHi - parent.longitudeLo) / lon_step)

    children = []
    for i in range(lat_count):
        for j in range(lon_count):
            lat = parent.latitudeLo + i * lat_step + lat_step / 2
            lon = parent.longitudeLo + j * lon_step + lon_step / 2
            child = encode(lat, lon, resolution)
            children.append(child)

    return children



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""


## WGS84 lat/long coordinate with an accuracy in metres squared.
class LatLongPoint(object):
    ## Sets the latitude, longitude and accuracy of the point.
    #  @param latitude Latitude of the point (in degrees).
    #  @param longitude Longitude of the point (in degrees).
    #  @param accuracy Accuracy of the point (in metres squared).
    def __init__(self, latitude, longitude, accuracy):
        self._latitude = latitude
        self._longitude = longitude
        self._accuracy = accuracy

    ## @return Latitude of the point (in degrees).
    def get_latitude(self):
        return self._latitude

    ## @return Longitude of the point (in degrees).
    def get_longitude(self):
        return self._longitude

    ## @return Accuracy of the point (in metres squared).
    def get_accuracy(self):
        return self._accuracy

    ## @return True if points have the same latitude, longitude and accuracy values,
    #          otherwise false.
    def __eq__(self, other):
        return (
            self._latitude == other.get_latitude()
            and self._longitude == other.get_longitude()
            and self._accuracy == other.get_accuracy()
        )

    ## @return Opposite of the == operator.
    def __ne__(self, other):
        return not (self == other)



import os
import argparse
import json
import csv
from tqdm import tqdm
import rasterio
from vgrid.utils import s2
import numpy as np
from shapely.geometry import Polygon
from vgrid.stats.s2stats import s2_metrics
from vgrid.utils.antimeridian import fix_polygon
from vgrid.generator.settings import geodesic_dggs_metrics, geodesic_dggs_to_feature
from math import cos, radians


def get_nearest_s2_resolution(raster_path):
    with rasterio.open(raster_path) as src:
        transform = src.transform
        crs = src.crs
        pixel_width = transform.a
        pixel_height = -transform.e
        cell_size = pixel_width * pixel_height

        if crs.is_geographic:
            # Latitude of the raster center
            center_latitude = (src.bounds.top + src.bounds.bottom) / 2
            # Convert degrees to meters
            meter_per_degree_lat = 111_320  # Roughly 1 degree latitude in meters
            meter_per_degree_lon = meter_per_degree_lat * cos(radians(center_latitude))

            pixel_width_m = pixel_width * meter_per_degree_lon
            pixel_height_m = pixel_height * meter_per_degree_lat
            cell_size = pixel_width_m * pixel_height_m

    nearest_resolution = None
    min_diff = float("inf")

    # Check resolutions from 0 to 24
    for res in range(25):
        _, _, avg_area = s2_metrics(res)
        diff = abs(avg_area - cell_size)
        # If the difference is smaller than the current minimum, update the nearest resolution
        if diff < min_diff:
            min_diff = diff
            nearest_resolution = res

    return nearest_resolution


def convert_numpy_types(obj):
    """Recursively convert NumPy types to native Python types"""
    if isinstance(obj, np.generic):
        return obj.item()  # Convert numpy types like np.uint8 to native Python int
    elif isinstance(obj, dict):
        return {k: convert_numpy_types(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_numpy_types(v) for v in obj]
    else:
        return obj


def raster2s2(raster_path, resolution=None, format="geojson"):
    # Step 1: Determine the nearest s2 resolution if none is provided
    if resolution is None:
        resolution = get_nearest_s2_resolution(raster_path)
        print(f"Nearest s2 resolution determined: {resolution}")

    # Open the raster file to get metadata and data
    with rasterio.open(raster_path) as src:
        raster_data = src.read()  # Read all bands
        transform = src.transform
        width, height = src.width, src.height
        band_count = src.count  # Number of bands in the raster

    s2_tokens = set()

    for row in range(height):
        for col in range(width):
            lon, lat = transform * (col, row)
            lat_lng = s2.LatLng.from_degrees(lat, lon)
            s2_id = s2.CellId.from_lat_lng(lat_lng)  # return S2 cell at max level 30
            s2_id = s2_id.parent(resolution)  # get S2 cell at resolution
            s2_token = s2.CellId.to_token(
                s2_id
            )  # get Cell ID Token, shorter than cell_id.id()
            s2_tokens.add(s2_token)

    # Sample the raster values at the centroids of the s2 hexagons
    s2_data = []

    for s2_token in tqdm(s2_tokens, desc="Resampling", unit=" cells"):
        cell_id = s2.CellId.from_token(s2_token)
        s2_cell = s2.Cell(cell_id)
        vertices = [s2_cell.get_vertex(i) for i in range(4)]

        shapely_vertices = []
        for vertex in vertices:
            lat_lng = s2.LatLng.from_point(vertex)  # Convert Point to LatLng
            longitude = lat_lng.lng().degrees  # Access longitude in degrees
            latitude = lat_lng.lat().degrees  # Access latitude in degrees
            shapely_vertices.append((longitude, latitude))

        # Close the polygon by adding the first vertex again
        shapely_vertices.append(shapely_vertices[0])  # Closing the polygon
        # Create a Shapely Polygon
        cell_polygon = fix_polygon(Polygon(shapely_vertices))  # Fix antimeridian
        num_edges = 4
        centroid_lat, centroid_lon, _, _ = geodesic_dggs_metrics(
            cell_polygon, num_edges
        )

        col, row = ~transform * (centroid_lon, centroid_lat)

        if 0 <= col < width and 0 <= row < height:
            # Get the values for all bands at this centroid
            values = raster_data[:, int(row), int(col)]
            s2_data.append(
                {
                    "s2": s2_token,
                    **{
                        f"band_{i + 1}": values[i] for i in range(band_count)
                    },  # Create separate columns for each band
                }
            )

    if format.lower() == "csv":
        import io

        output = io.StringIO()
        if s2_data:
            writer = csv.DictWriter(output, fieldnames=s2_data[0].keys())
            writer.writeheader()
            writer.writerows(s2_data)
        return output.getvalue()

    # Create the GeoJSON-like structure
    s2_features = []
    for data in tqdm(s2_data, desc="Converting to GeoJSON", unit=" cells"):
        cell_id = s2.CellId.from_token(data["s2"])
        s2_cell = s2.Cell(cell_id)
        if s2_cell:
            # Get the vertices of the cell (4 vertices for a rectangular cell)
            vertices = [s2_cell.get_vertex(i) for i in range(4)]
            # Prepare vertices in (longitude, latitude) format for Shapely
            shapely_vertices = []
            for vertex in vertices:
                lat_lng = s2.LatLng.from_point(vertex)  # Convert Point to LatLng
                longitude = lat_lng.lng().degrees  # Access longitude in degrees
                latitude = lat_lng.lat().degrees  # Access latitude in degrees
                shapely_vertices.append((longitude, latitude))

            # Close the polygon by adding the first vertex again
            shapely_vertices.append(shapely_vertices[0])  # Closing the polygon
            # Create a Shapely Polygon
            cell_polygon = fix_polygon(Polygon(shapely_vertices))  # Fix antimeridian
            num_edges = 4
            band_properties = {
                f"band_{i + 1}": data[f"band_{i + 1}"] for i in range(band_count)
            }
            s2_feature = geodesic_dggs_to_feature(
                "s2", data["s2"], resolution, cell_polygon, num_edges
            )
            s2_feature["properties"].update(convert_numpy_types(band_properties))

            s2_features.append(s2_feature)

    return {"type": "FeatureCollection", "features": s2_features}


def raster2s2_cli():
    parser = argparse.ArgumentParser(
        description="Convert Raster in Geographic CRS to S2 DGGS"
    )
    parser.add_argument("-raster", type=str, required=True, help="Raster file path")

    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        required=False,
        default=None,
        help="Resolution [0..24]",
    )

    parser.add_argument(
        "-f",
        "--format",
        type=str,
        required=False,
        default="geojson",
        choices=["geojson", "csv"],
        help="Output format (geojson or csv)",
    )

    args = parser.parse_args()
    raster = args.raster
    resolution = args.resolution
    format = args.format

    if not os.path.exists(raster):
        print(f"Error: The file {raster} does not exist.")
        return

    if resolution is not None:
        if resolution < 0 or resolution > 24:
            print("Please select a resolution in [0..24] range and try again")
            return

    # Process the raster
    result = raster2s2(raster, resolution, format)

    # Generate output filename
    base_name = os.path.splitext(os.path.basename(raster))[0]
    output_path = f"{base_name}2s2.{format}"

    # Save the output
    if format.lower() == "geojson":
        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(result, f)
    elif format.lower() == "csv":
        fieldnames = list(result[0].keys())
        with open(output_path, "w", newline="", encoding="utf-8") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(result)

    print(f"Output saved as {output_path}")



import platform
import argparse
import os
import json
import statistics
from collections import defaultdict, Counter
from shapely.geometry import Point, Polygon
from shapely.wkt import loads

from tqdm import tqdm
from vgrid.binning.bin_helper import get_default_stats_structure, append_stats_value
from vgrid.generator.settings import geodesic_dggs_to_feature

if platform.system() == "Windows":
    from vgrid.utils.eaggr.enums.shape_string_format import ShapeStringFormat
    from vgrid.utils.eaggr.eaggr import Eaggr
    from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
    from vgrid.utils.eaggr.enums.model import Model
    from vgrid.generator.isea4tgrid import fix_isea4t_wkt, fix_isea4t_antimeridian_cells
    from vgrid.conversion.latlon2dggs import latlon2isea4t


def isea4t_bin(isea4t_dggs, point_features, resolution, stats, category, field_name):
    if platform.system() == "Windows":
        isea4t_bins = defaultdict(lambda: defaultdict(get_default_stats_structure))

        for point_feature in tqdm(point_features, desc="Binning points"):
            geom = point_feature["geometry"]
            props = point_feature.get("properties", {})

            if geom["type"] == "Point":
                point = Point(geom["coordinates"])
                isea4t_id = latlon2isea4t(point.y, point.x, resolution)
                append_stats_value(
                    isea4t_bins, isea4t_id, props, stats, category, field_name
                )

            elif geom["type"] == "MultiPoint":
                for coords in geom["coordinates"]:
                    point = Point(coords)
                    isea4t_id = latlon2isea4t(point.y, point.x, resolution)
                    append_stats_value(
                        isea4t_bins, isea4t_id, props, stats, category, field_name
                    )

        isea4t_features = []
        for isea4t_id, categories in isea4t_bins.items():
            cell_to_shape = isea4t_dggs.convert_dggs_cell_outline_to_shape_string(
                DggsCell(isea4t_id), ShapeStringFormat.WKT
            )
            cell_to_shape_fixed = loads(fix_isea4t_wkt(cell_to_shape))
            if (
                isea4t_id.startswith("00")
                or isea4t_id.startswith("09")
                or isea4t_id.startswith("14")
                or isea4t_id.startswith("04")
                or isea4t_id.startswith("19")
            ):
                cell_to_shape_fixed = fix_isea4t_antimeridian_cells(cell_to_shape_fixed)

            cell_polygon = Polygon(list(cell_to_shape_fixed.exterior.coords))
            num_edges = 3

            if not cell_polygon.is_valid:
                continue

            isea4t_feature = geodesic_dggs_to_feature(
                "isea4t", isea4t_id, resolution, cell_polygon, num_edges
            )

            for cat, values in categories.items():
                key_prefix = "" if category is None else f"{cat}_"

                if stats == "count":
                    isea4t_feature["properties"][f"{key_prefix}count"] = values["count"]
                elif stats == "sum":
                    isea4t_feature["properties"][f"{key_prefix}sum"] = sum(
                        values["sum"]
                    )
                elif stats == "min":
                    isea4t_feature["properties"][f"{key_prefix}min"] = min(
                        values["min"]
                    )
                elif stats == "max":
                    isea4t_feature["properties"][f"{key_prefix}max"] = max(
                        values["max"]
                    )
                elif stats == "mean":
                    isea4t_feature["properties"][f"{key_prefix}mean"] = statistics.mean(
                        values["mean"]
                    )
                elif stats == "median":
                    isea4t_feature["properties"][f"{key_prefix}median"] = (
                        statistics.median(values["median"])
                    )
                elif stats == "std":
                    isea4t_feature["properties"][f"{key_prefix}std"] = (
                        statistics.stdev(values["std"]) if len(values["std"]) > 1 else 0
                    )
                elif stats == "var":
                    isea4t_feature["properties"][f"{key_prefix}var"] = (
                        statistics.variance(values["var"])
                        if len(values["var"]) > 1
                        else 0
                    )
                elif stats == "range":
                    isea4t_feature["properties"][f"{key_prefix}range"] = (
                        max(values["range"]) - min(values["range"])
                        if values["range"]
                        else 0
                    )

                elif stats == "minority":
                    freq = Counter(values["values"])
                    min_item = (
                        min(freq.items(), key=lambda x: x[1])[0] if freq else None
                    )
                    isea4t_feature["properties"][f"{key_prefix}minority"] = min_item

                elif stats == "majority":
                    freq = Counter(values["values"])
                    max_item = (
                        max(freq.items(), key=lambda x: x[1])[0] if freq else None
                    )
                    isea4t_feature["properties"][f"{key_prefix}majority"] = max_item

                elif stats == "variety":
                    isea4t_feature["properties"][f"{key_prefix}variety"] = len(
                        set(values["values"])
                    )

            isea4t_features.append(isea4t_feature)

        return isea4t_features


def main():
    parser = argparse.ArgumentParser(description="Binning point to isea4t DGGS")
    parser.add_argument(
        "-point",
        "--point",
        type=str,
        required=True,
        help="GeoJSON file path (Point or MultiPoint)",
    )
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        default=13,
        help="Resolution of the grid [0..25]",
    )
    parser.add_argument(
        "-stats",
        "--statistics",
        choices=[
            "count",
            "min",
            "max",
            "sum",
            "mean",
            "median",
            "std",
            "var",
            "range",
            "minority",
            "majority",
            "variety",
        ],
        required=True,
        help="Statistic option",
    )

    parser.add_argument(
        "-category",
        "--category",
        required=False,
        help="Optional category field for grouping",
    )
    parser.add_argument(
        "-field", "--field", required=False, help="Numeric field to compute statistics"
    )

    if platform.system() == "Windows":
        isea4t_dggs = Eaggr(Model.ISEA4T)
        args = parser.parse_args()

        resolution = args.resolution
        point = args.point
        stats = args.statistics
        category = args.category
        field_name = args.field

        if resolution < 0 or resolution > 25:
            print("Error: Please select a resolution in [0..25].")
            return

        if not os.path.exists(point):
            print(f"Error: The file {point} does not exist.")
            return

        if stats != "count" and not field_name:
            print("Error: A field name is required for statistics other than 'count'.")
            return

        with open(point, "r", encoding="utf-8") as f:
            point_data = json.load(f)

        point_features = point_data["features"]
        isea4t_features = isea4t_bin(
            isea4t_dggs, point_features, resolution, stats, category, field_name
        )

        out_name = os.path.splitext(os.path.basename(point))[0]
        out_path = f"{out_name}_bin_isea4t_{resolution}_{stats}.geojson"

        with open(out_path, "w", encoding="utf-8") as f:
            json.dump(
                {"type": "FeatureCollection", "features": isea4t_features}, f, indent=2
            )

        print(f"GeoJSON saved as {out_path}")


if __name__ == "__main__":
    main()



"""
Module for formatting responses for API.

© Regents of the University of Minnesota. All rights reserved.
This software is released under an Apache 2.0 license. Further details about the Apache 2.0 license are available in the license.txt file.
"""

from itertools import chain, product, tee

import numpy as np
import geopandas as gpd

from shapely import wkt
from shapely.geometry import Polygon, Point

from vgrid.utils.easedggs.constants import levels_specs, ease_crs, grid_spec
from vgrid.utils.easedggs.logConfig import logger


def format_response(data, success):
    """
    Format the library repsonse for return to API
    Args:
         data: list
            Contains data, or the error message
        success :(boolean)
            Indicatication of function success or failure.
    Returns:
        response : dict
            The formatted response to return to the API.
    """
    if success:
        response = {"success": success, "result": {"data": data}}

    else:
        response = {"success": success, "result": {"error_message": data}}

    return response


def boundbox_to_poly(left, bottom, right, top):
    """
    convert bounding extents to polygon
    Parameters
    ----------
    left : float
        Left edge of bounding box.
    bottom : float
        Bottom edge of bounding box.
    right : float
        Right edge of bounding box.
    top : TYPE
        Top edge of bounding box.

    Returns
    -------
    Shapely Polygon corresponding with bounding box.
    """
    if type(left) is int:
        left = float(left)
        bottom = float(bottom)
        right = float(right)
        top = float(top)

    ul = Point(left, top)
    ur = Point(right, top)
    lr = Point(right, bottom)
    ll = Point(left, bottom)

    poly = Polygon([ul, ur, lr, ll, ul])

    return poly


def get_polygon_corners(polygon, ccw=True):
    """
    Get the bounds of a polygon

    Parameters
    ----------
    polygon : shapely polygon
        The polygon to get the corner coordinates for.
    ccw : boolean
        Return coordinates in counter-clockwise order. Default = False

    Returns
    -------
    bounds : tuple
        Corner coodinates for the bounding box/evevlope of the polygon.
    """

    if type(polygon) is str:
        polygon = wkt.loads(polygon)

    # .bounds method returns (minx, miny, maxx, maxy) by default.
    #    that is the same as (left, bottom, right, top).
    #  Keep that convetion, but allow for return in clockwise order.
    #     when false: (left, top, right, bottom) or: (minx, maxy, maxx, miny)
    minx, miny, maxx, maxy = polygon.bounds

    upper_left = Point(minx, maxy)
    upper_right = Point(maxx, maxy)
    lower_left = Point(minx, miny)
    lower_right = Point(maxx, miny)

    if not ccw:
        return upper_left, lower_left, lower_right, upper_right

    return upper_left, upper_right, lower_right, lower_left


def enumerate_id_elements(refine_ratio):
    if not isinstance(refine_ratio, int):
        return False

    row_ids = [str(rr) for rr in range(refine_ratio)]
    col_ids = row_ids

    sub_elements = list(chain(product(row_ids, col_ids)))
    sub_elements = ["".join(se) for se in sub_elements]

    return sub_elements


def enumerate_grid_table_rows(x_coords, y_coords, level=0, parent_id=None):
    """
    Generate grid polygons for cells using x, y cooridnate vectors.
    Parameters
    ----------
    x_coords : float
       X coordinate vector in EASE grid coordinates.
    y_coords : float
        Y coordinate vector in EASE grid coordinates.
    level _ int
        The grid level for the corresponding geometries.
    parent_id : String
        Denotes the cell id of the parent cell. Default is None.

    Returns
    -------
    Relevant child cell characteritics: Row ID, Column ID, Grid ID, Geometry and Centroid
    """
    geoms = []
    r_ind = []
    c_ind = []
    grid_ids = []
    centroid = []
    if level == 0:
        pass
    #         children = []

    np.arange(1, len(x_coords))
    np.arange(1, len(y_coords))

    for r, c in product(np.arange(1, len(y_coords), 1), np.arange(1, len(x_coords), 1)):
        c = int(c)
        r = int(r)

        poly = boundbox_to_poly(
            left=x_coords[c - 1],
            bottom=y_coords[r],
            right=x_coords[c],
            top=y_coords[r - 1],
        )
        geoms.append(poly.wkt)
        centroid.append(poly.centroid.wkt)
        ci = c - 1
        ri = r - 1

        r_ind.append(ri)
        c_ind.append(ci)

        if level == 0:
            grid_id = "L0.{:03d}{:03d}".format(ri, ci)

        else:
            grid_id = "{}.{:d}{:d}".format(parent_id, ri, ci)

        grid_ids.append(grid_id)

    return r_ind, c_ind, grid_ids, geoms, centroid


def gen_point_grid(
    upper_left,
    upper_right,
    lower_right,
    lower_left,
    delta_x,
    delta_y,
    target_crs=ease_crs,
):
    """
    Generates a grid of equally spaced points, between geograhic extent

    upper_left : shapely Point
        cooridinate for the upper left of the geographic extent

    upper_right : shapely Point
        cooridinate for the upper right of the geographic extent

    lower_right : shapely Point
        cooridinate for the lower right of the geographic extent

    lower_left : shapely Point
        cooridnate for the lower left of the geographic extent

    delta_x : float
        horizontal distance between subsequent cooridnates

    delta_y : float
        vertical distance between subsequent cooridnates

    returns:
    point_grid : pandas GeoDataFame
        grid of evenly spaced coordinates between the geographic extent
    """
    if not (
        isinstance(upper_left, Point)
        and isinstance(upper_right, Point)
        and isinstance(lower_right, Point)
        and isinstance(lower_left, Point)
    ):
        return False

    if not (isinstance(delta_x, float) and isinstance(delta_y, float)):
        return False

    maxx = np.max([upper_left.x, upper_right.x, lower_right.x, lower_left.x])
    minx = np.min([upper_left.x, upper_right.x, lower_right.x, lower_left.x])
    # print(maxx, minx)
    logger.debug("gen_point_grid: maxx - {}; minx - {}".format(maxx, minx))

    maxy = np.max([upper_left.y, upper_right.y, lower_right.y, lower_left.y])
    miny = np.min([upper_left.y, upper_right.y, lower_right.y, lower_left.y])

    x_steps = np.rint((maxx - minx) / delta_x).astype("int") + 1
    y_steps = np.rint((maxy - miny) / delta_y).astype("int") + 1

    x_coords = np.linspace(start=minx, stop=maxx, num=x_steps, endpoint=True)
    y_coords = np.linspace(start=maxy, stop=miny, num=y_steps, endpoint=True)

    grid = np.meshgrid(x_coords, y_coords, indexing="xy")

    x_points = np.ravel(grid[0])
    y_points = np.ravel(grid[1])

    coords = [Point(x_points[i], y_points[i]) for i in range(x_points.shape[0])]

    point_grid = gpd.GeoDataFrame(
        {"ease_x": x_points, "ease_y": y_points, "geometry": coords}, crs=target_crs
    )

    return point_grid


def calc_grid_coord_vectors(
    min_x,
    min_y,
    max_x,
    max_y,
    level=0,
    levels_specs=levels_specs,
    x_ascend=True,
    y_ascend=False,
):
    """
    Calculate pixel corner coordinates for the specified level using grid bounding coordinates

    Parameters
    ----------
    min_x : float
        EASE Grid cooridinates of left edge of bounding box.
    min_y : float
        EASE Grid cooridinates of bottom edge of bounding box.
    max_x : float
        EASE Grid cooridinates of right edge of bounding box.
    max_y : TYPE
        EASE Grid cooridinates of top edge of bounding box.

    Returns
    -------
    x_coord_vector, y_coord_vector : np.array
        Numpy arrays of x, y cooridinates at specified level.
    """
    # user in puts the number of columns/rows of the grid.
    # number of grid edges is column/rows +1
    #
    # for level 0, coordinates are generated using the number of rows/columns for the grid
    if level == 0:
        x_col = levels_specs[level]["n_col"] + 1
        y_row = levels_specs[level]["n_row"] + 1

    # for all levels > 0, the refine_ratio is used to derive the number of cells.
    #    the refine_ratio for levels is associuated with the partent in the dict
    else:
        level = level - 1
        level = str(level)
        x_col = levels_specs[level]["refine_ratio"] + 1
        y_row = levels_specs[level]["refine_ratio"] + 1

    if x_ascend:
        x_coord_vector = np.linspace(start=min_x, stop=max_x, num=x_col)
    else:
        x_coord_vector = np.linspace(start=max_y, stop=min_x, num=x_col)

    if y_ascend:
        y_coord_vector = np.linspace(start=min_y, stop=max_y, num=y_row)
    else:
        y_coord_vector = np.linspace(start=max_y, stop=min_y, num=y_row)

    return x_coord_vector, y_coord_vector


######################
#
# helper function declarations
#
######################
# adapted from: https://math.stackexchange.com/questions/914823/shift-numbers-into-a-different-range
def _shift_range(val, o_start, o_end, n_start, n_end):
    """Map value (x) from old range to new range of values.

    Parameters
    ----------
    val : numeric
        Specific numeric value to shift
    o_start : numeric
        Start value of original range
    o_end : numeric
        End value of original range
    n_start : numeric
        Start value of new range
    n_end : _type_
        End value of new range

    Returns
    -------
    numeric
        Value mapped into new range
    """
    return (n_end - n_start) / (o_end - o_start) * (val - o_start) + n_start


def shift_range_ease(val, axis):
    """
    Convert 1-d EASE v2 coordinate (x | y) to EASE-DGGS coordinate.

    Parameters
    ----------
    val: float
        The value to shift from EASE coords to EASE-DGGS coords.
    axis: str
        'x' or 'y'; the axis of the coords to be shifted.

        (we should be able to use an enum and pydantic or something
        to declare an acceptable set of values)

    Returns
    -------
    numeric
        Value mapped into EASE v2 bounding box range.
    """
    if axis == "x":
        return _shift_range(
            val,
            grid_spec["ease"][f"min_{axis}"],
            grid_spec["ease"][f"max_{axis}"],
            0,
            levels_specs[0]["n_col"],
        )

    if axis == "y":
        return _shift_range(
            val,
            grid_spec["ease"][f"max_{axis}"],
            grid_spec["ease"][f"min_{axis}"],
            0,
            levels_specs[0]["n_row"],
        )


def shift_range_grid_xy(val, axis):
    """
    Convert 1-d EASE-DGGS coordinate (x | y) to EASE v2 grid coordinate.

    Parameters
    ----------
    val: float
        The value to shift from EASE coords to EASE-DGGS coords.
    axis: str
        'x' or 'y'; the axis of the coords to be shifted.

        (we should be able to use an enum and pydantic or something
        to declare an acceptable set of values)
    Returns
    -------
    numeric
        Value mapped into EASE-DGGS space (x|cols [0: 964]; y|rows [0: 406])
    """
    if axis == "x":
        return _shift_range(
            val,
            0,
            levels_specs[0]["n_col"],
            grid_spec["ease"][f"min_{axis}"],
            grid_spec["ease"][f"max_{axis}"],
        )
    if axis == "y":
        return _shift_range(
            val,
            0,
            levels_specs[0]["n_row"],
            grid_spec["ease"][f"max_{axis}"],
            grid_spec["ease"][f"min_{axis}"],
        )


def shift_range_grid_multiple(val, axis):
    """
    Convert 1-d EASE-DGGS coordinate (x | y) * levels_specs[0]['x_lenth']
    to EASE v2 grid coordinate.

    Parameters
    ----------
    val: float
        The value to shift from EASE coords to EASE-DGGS coords.
    axis: str
        'x' or 'y'; the axis of the coords to be shifted.

        (we should be able to use an enum and pydantic or something to
        declare an acceptable set of values)

    Returns
    -------
    numeric
        Value mapped from (n_rows | n_cols) * cell_widths to EASE v2 bounds of
        EASE-DGGS (min|max x|y)
    """
    if axis == "x":
        return _shift_range(
            val,
            0,
            levels_specs[0]["n_col"] * levels_specs[0]["x_length"],
            grid_spec["ease"][f"min_{axis}"],
            grid_spec["ease"][f"max_{axis}"],
        )
    if axis == "y":
        return _shift_range(
            val,
            0,
            levels_specs[0]["n_row"] * levels_specs[0]["y_length"],
            grid_spec["ease"][f"max_{axis}"],
            grid_spec["ease"][f"min_{axis}"],
        )


# truncate function from:
#   https://stackoverflow.com/questions/42021972/truncating-decimal-digits-numpy-array-of-floats
def trunc(vals, decimals=0):
    """Take values and truncates them at specific decimal place

    Parameters
    ----------
    values : np.array, floats
        Floats to truncate at decimal place.
    decimals : int, optional
        The decimal place to truncate the , by default 0

    Returns
    -------
    numpy.array
        Array truncated at deimal place
    """
    return np.trunc(vals * 10**decimals) / (10**decimals)


def flatten(list):
    """Function that converts nested lists into a flat list.

    Parameters
    ----------
    list : list
        Nested list to convert to individual elements

    Returns
    -------
    flattened list
        Single list indvidual elements
    """
    return [li for sub_list in list for li in sub_list]


# often end up doing operations going around in a 'circle', particularly
#   operations around the pairs of a bounding box. the pairwise_circle
#   is not a new idea, and hence from stack exchange
# https://stackoverflow.com/questions/36917042/pairwise-circular-python-for-loop/36927946#36927946
def pairwise_circle(iterable, reverse=False):
    """Creates pairwise combinations from a list, with last element looping back to first.
        E.g. [UL, UR, LR, LL] -> [(UL, UR), (UR, LR), (LR, LL), (LL, UL)]
        if reverse = True
        E.g. [UL, UR, LR, LL] -> [(LL, LR), (LR, UR), (UR, UL), (UL, LL)]

    Parameters
    ----------
    iterable : _type_
        The iterable to construct pairwise combinations of.
    reverse : bool, optional
        Perform pair wise combinations in reverse order, by default False


    Returns
    -------
    iterable
        Tuples of the pairwise combinations
    """
    if reverse:
        iterable = reversed(iterable)

    a, b = tee(iterable)
    first = next(b, None)
    return zip(a, chain(b, (first,)))


def epsilon_check(x, y, E=1e-5) -> bool:
    """Check that difference between 2 values is less than EPISLON.

    Parameters
    ----------
    x : float
        First value to compare
    y : float
        Second value to compare
    E : float, optional
        The value the difference between x, y should not exceed, by default EPSILON

    Returns
    -------
    bool
        True ff difference between x, y is less than, equal to EPSILON
    """
    return abs(x - y) <= E

    # basic idea of add_node is that a line segement can be minimally defined


#   using two points. here, these are teh start, end points. The aim of this
#   function is to add addition points into the line segment at regular intervals
#   condider:
#
#   initial condition:
#   start                    end
#   O------------------------O
#
#   after add_nodes nodes =7
#   st  n1  n2  n3  n4  n5  n6  n7  ed
#   O---o---o---o---o---o---o---o---O
def add_nodes(start, end, nodes=21):
    """Insert nodes (coordinate pairs) between the start, end of a line segment.

    Parameters
    ----------
    start : tuple
        Starting coordinate pair of line segment to add nodes to.
    end : tuple
        Ending coordinate pair of line segment to add nodes to.
    nodes : int, optional
        Number of nodes to add to line segement, by default 21.
    # drop_end : boolean, optional
    #     Drop the final coordinate pair of the series, by default True.
    Returns
    ----------
    new line segement : list
        List with nodes added inbetween start, end.
    """
    xs = np.linspace(start[0], end[0], nodes)
    ys = np.linspace(start[1], end[1], nodes)

    return list(zip(xs[:], ys[:]))
